"""
TODO:
 - CancellableFuture
"""
import abc
import typing as ta

import anyio

from ... import lang


T = ta.TypeVar('T')


##


class FutureError(Exception):
    pass


class FutureOutcomeAlreadySetError(FutureError):
    pass


##


class Future(lang.Abstract, ta.Awaitable[lang.Outcome[T]]):
    @abc.abstractmethod
    def __await__(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def outcome(self) -> lang.Maybe[lang.Outcome[T]]:
        raise NotImplementedError

    @abc.abstractmethod
    def set_outcome(self, o: lang.Outcome[T]) -> None:
        raise NotImplementedError

    def set_value(self, v: T) -> None:
        self.set_outcome(lang.value(v))

    def set_error(self, e: BaseException) -> None:
        self.set_outcome(lang.error(e))


##


class _FutureImpl(Future[T]):
    def __init__(self, *, event: anyio.Event | None = None) -> None:
        super().__init__()

        self._outcome: lang.Maybe[lang.Outcome[T]] = lang.empty()

        if event is None:
            event = anyio.Event()
        self._event = event

    def __await__(self):
        if (o := self._outcome).present:
            return o
        yield from self._event.wait().__await__()
        return self._outcome.must()

    @property
    def outcome(self) -> lang.Maybe[lang.Outcome[T]]:
        return self._outcome

    def set_outcome(self, o: lang.Outcome[T]) -> None:
        if self._outcome.present:
            raise FutureOutcomeAlreadySetError
        self._outcome = lang.just(o)
        self._event.set()


##


def _create_future() -> Future[T]:
    return _FutureImpl()


# PEP695 workaround
class create_future(Future[T]):  # noqa
    def __new__(cls) -> Future[T]:  # type: ignore[misc]
        return _create_future()

    def __init__(self) -> None:
        raise TypeError

    def __await__(self):
        raise TypeError

    @property
    def outcome(self) -> lang.Maybe[lang.Outcome[T]]:
        raise TypeError

    def set_outcome(self, o: lang.Outcome[T]) -> None:
        raise TypeError



# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import functools
import operator
import typing as ta


T = ta.TypeVar('T')


##


def merge(seqs: ta.MutableSequence[list[T]]) -> list[T]:
    """
    Merges MROs in *sequences* to a single MRO using the C3 algorithm.

    Adapted from https://www.python.org/download/releases/2.3/mro/.
    """

    result: list[T] = []
    candidate: T | None = None
    while True:
        seqs = [s for s in seqs if s]  # purge empty sequences
        if not seqs:
            return result
        for s1 in seqs:  # find merge candidates among seq heads
            candidate = s1[0]
            for s2 in seqs:
                if candidate in s2[1:]:
                    candidate = None
                    break  # reject the current head, it appears later
            else:
                break
        if candidate is None:
            raise RuntimeError('Inconsistent hierarchy')
        result.append(candidate)
        # remove the chosen candidate
        for seq in seqs:
            if seq[0] == candidate:
                del seq[0]


def _default_is_abstract(obj: ta.Any) -> bool:
    return hasattr(obj, '__abstractmethods__')


def mro(
        cls: T,
        abcs: ta.Sequence[T] | None = None,
        *,
        get_bases: ta.Callable[[T], ta.Sequence[T]] = operator.attrgetter('__bases__'),
        is_subclass: ta.Callable[[T, T], bool] = issubclass,  # type: ignore
        is_abstract: ta.Callable[[T], bool] = _default_is_abstract,
) -> list[T]:
    """
    Computes the method resolution order using extended C3 linearization.

    If no *abcs* are given, the algorithm works exactly like the built-in C3 linearization used for method resolution.

    If given, *abcs* is a list of abstract base classes that should be inserted into the resulting MRO. Unrelated ABCs
    are ignored and don't end up in the result. The algorithm inserts ABCs where their functionality is introduced, i.e.
    issubclass(cls, abc) returns True for the class itself but returns False for all its direct base classes. Implicit
    ABCs for a given class (either registered or inferred from the presence of a special method like __len__) are
    inserted directly after the last ABC explicitly listed in the MRO of said class. If two implicit ABCs end up next to
    each other in the resulting MRO, their ordering depends on the order of types in *abcs*.
    """

    for i, base in enumerate(reversed(get_bases(cls))):
        if is_abstract(base):
            boundary = len(get_bases(cls)) - i
            break  # Bases up to the last explicit ABC are considered first.
    else:
        boundary = 0
    abcs = list(abcs) if abcs else []
    explicit_bases = list(get_bases(cls)[:boundary])
    abstract_bases = []
    other_bases = list(get_bases(cls)[boundary:])
    for base in abcs:
        if (
                is_subclass(cls, base) and  # noqa
                not any(is_subclass(b, base) for b in get_bases(cls))  # noqa
        ):
            # If *cls* is the class that introduces behaviour described by an ABC *base*, insert said ABC to its MRO.
            abstract_bases.append(base)
    for base in abstract_bases:
        abcs.remove(base)
    rec = functools.partial(
        mro,
        abcs=abcs,
        get_bases=get_bases,
        is_subclass=is_subclass,
        is_abstract=is_abstract,
    )
    explicit_c3_mros = [rec(base) for base in explicit_bases]
    abstract_c3_mros = [rec(base) for base in abstract_bases]
    other_c3_mros = [rec(base) for base in other_bases]
    return merge(
        [[cls]] +  # noqa
        explicit_c3_mros + abstract_c3_mros + other_c3_mros +
        [explicit_bases] + [abstract_bases] + [other_bases],
    )


def compose_mro(
        cls: T,
        types: ta.Iterable[T],
        *,
        get_mro: ta.Callable[[T], ta.Sequence[T] | None] = operator.attrgetter('__mro__'),
        get_bases: ta.Callable[[T], ta.Sequence[T]] = operator.attrgetter('__bases__'),
        is_subclass: ta.Callable[[T, T], bool] = issubclass,  # type: ignore
        get_subclasses: ta.Callable[[T], ta.Iterable[T]] = operator.methodcaller('__subclasses__'),
        is_abstract: ta.Callable[[T], bool] = _default_is_abstract,
) -> list[T]:
    """
    Calculates the method resolution order for a given class *cls*.

    Includes relevant abstract base classes (with their respective bases) from the *types* list. Uses a modified C3
    linearization algorithm.
    """

    bases = set(get_mro(cls) or [])

    # Remove entries which are already present in the __mro__ or unrelated.
    def is_related(typ):
        return typ not in bases and get_mro(typ) is not None and is_subclass(cls, typ)

    types = [n for n in types if is_related(n)]

    # Remove entries which are strict bases of other entries (they will end up in the MRO anyway.
    def is_strict_base(typ):
        for other in types:  # noqa
            if typ != other and typ in (get_mro(other) or []):
                return True
        return False

    types = [n for n in types if not is_strict_base(n)]

    # Subclasses of the ABCs in *types* which are also implemented by *cls* can be used to stabilize ABC ordering.
    type_set = set(types)
    _mro = []
    for typ in types:
        found: list[list[T]] = []
        for sub in get_subclasses(typ):
            if sub not in bases and is_subclass(cls, sub):  # noqa
                found.append([s for s in (get_mro(sub) or []) if s in type_set])
        if not found:
            _mro.append(typ)
            continue

        # Favor subclasses with the biggest number of useful bases
        found.sort(key=len, reverse=True)
        for lst in found:
            for subcls in lst:
                if subcls not in _mro:
                    _mro.append(subcls)

    return mro(
        cls,
        abcs=_mro,
        get_bases=get_bases,
        is_subclass=is_subclass,
        is_abstract=is_abstract,
    )



import dataclasses as dc
import typing as ta

from .. import lang
from .api.classes.conversion import std_params_to_class_spec
from .api.classes.metadata import extract_cls_metadata
from .api.classes.params import get_class_spec
from .api.fields.conversion import std_field_to_field_spec
from .concerns.fields import InitFields
from .concerns.fields import calc_init_fields
from .inspect import FieldsInspection
from .inspect import inspect_fields
from .internals import STD_PARAMS_ATTR
from .internals import StdFieldType
from .internals import std_field_type
from .specs import ClassSpec


##


class ClassReflection:
    def __init__(self, cls: type) -> None:
        super().__init__()

        self._cls = cls

    @property
    def cls(self) -> type:
        return self._cls

    @lang.cached_property
    def spec(self) -> ClassSpec:
        if (cs := get_class_spec(self._cls)) is not None:
            return cs

        try:
            p = getattr(self._cls, STD_PARAMS_ATTR)
        except AttributeError:
            raise TypeError(self._cls) from None

        # This class was not constructed as one of our dataclasses, so surface no additional functionality even if
        # present as extra_params or such.
        fsl = [
            std_field_to_field_spec(
                f,
                ignore_metadata=True,
                ignore_extra_params=True,
            )
            for f in dc.fields(self._cls)  # noqa
        ]

        cmd = extract_cls_metadata(self._cls, deep=True)

        cs = std_params_to_class_spec(
            p,
            fsl,
            metadata=cmd.user_metadata or None,
        )

        return cs

    @lang.cached_property
    def fields_inspection(self) -> FieldsInspection:
        return inspect_fields(self._cls)

    @lang.cached_property
    def init_fields(self) -> InitFields:
        return calc_init_fields(
            self.spec.fields,
            reorder=self.spec.reorder,
            class_kw_only=self.spec.kw_only,
        )

    @lang.cached_property
    def fields(self) -> ta.Mapping[str, dc.Field]:
        return {f.name: f for f in dc.fields(self._cls)}  # noqa

    @lang.cached_property
    def instance_fields(self) -> ta.Sequence[dc.Field]:
        return [f for f in self.fields.values() if std_field_type(f) is StdFieldType.INSTANCE]


def reflect(cls: type) -> ClassReflection:
    return ClassReflection(cls)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
https://github.com/util-linux/util-linux/blob/a4436c7bf07f98a6381c7dfa2ab3f9a415f9c479/misc-utils/lslocks.c
"""
import dataclasses as dc
import json
import typing as ta

from ..lite.check import check
from ..lite.marshal import OBJ_MARSHALER_FIELD_KEY
from ..lite.marshal import unmarshal_obj
from ..subprocesses.run import SubprocessRun
from ..subprocesses.run import SubprocessRunnable
from ..subprocesses.run import SubprocessRunOutput


##


@dc.dataclass(frozen=True)
class LslocksItem:
    """https://manpages.ubuntu.com/manpages/lunar/man8/lslocks.8.html"""

    command: str
    pid: int
    type: str  # POSIX | FLOCK | OFDLCK
    size: ta.Optional[int]
    mode: str  # READ | WRITE
    mandatory: bool = dc.field(metadata={OBJ_MARSHALER_FIELD_KEY: 'm'})
    start: int
    end: int
    path: str
    blocker: ta.Optional[int] = None


##


@dc.dataclass(frozen=True)
class LslocksCommand(SubprocessRunnable[ta.List[LslocksItem]]):
    pid: ta.Optional[int] = None
    no_inaccessible: bool = False

    def make_run(self) -> SubprocessRun:
        return SubprocessRun.of(
            'lslocks',
            '--json',
            '--bytes',
            '--notruncate',
            *(['--pid', str(self.pid)] if self.pid is not None else []),
            *(['--noinaccessible'] if self.no_inaccessible else []),

            check=True,
            stdout='pipe',
            stderr='devnull',
        )

    def handle_run_output(self, output: SubprocessRunOutput) -> ta.List[LslocksItem]:
        buf = check.not_none(output.stdout).decode().strip()
        if not buf:
            return []
        obj = json.loads(buf)
        return unmarshal_obj(obj['locks'], ta.List[LslocksItem])



import dataclasses as dc
import typing as ta

from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.globals import REPRLIB_RECURSIVE_REPR_GLOBAL
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.ops import OpRef
from ..generation.ops import Ref
from ..generation.registry import register_generator_type
from ..processing.base import ProcessingContext
from ..specs import FieldSpec
from ..specs import FieldType
from ..specs import ReprFn
from .fields import InitFields


##


@dc.dataclass(frozen=True, kw_only=True)
class ReprPlan(Plan):
    @dc.dataclass(frozen=True, kw_only=True)
    class Field:
        name: str
        kw_only: bool = False
        fn: OpRef[ReprFn] | None = None

    fields: tuple[Field, ...]

    id: bool = False
    terse: bool = False


@register_generator_type(ReprPlan)
class ReprGenerator(Generator[ReprPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[ReprPlan] | None:
        if not ctx.cs.repr or '__repr__' in ctx.cls.__dict__:
            return None

        ifs = ctx[InitFields]
        fs: ta.Sequence[FieldSpec]
        if ctx.cs.terse_repr:
            # If terse repr will match init param order
            fs = ifs.all
        else:
            # Otherwise default to dc.fields() order
            fs = sorted(ctx.cs.fields, key=lambda f: f.repr_priority or 0)

        orm = {}
        rfs: list[ReprPlan.Field] = []
        for i, f in enumerate(fs):
            if not (f.field_type is FieldType.INSTANCE and f.repr):
                continue

            fnr: OpRef | None = None
            if f.repr_fn is not None:
                fnr = OpRef(f'repr.fns.{i}.fn')
                orm[fnr] = f.repr_fn

            rfs.append(ReprPlan.Field(
                name=f.name,
                kw_only=f in ifs.kw_only,
                fn=fnr,
            ))

        return PlanResult(
            ReprPlan(
                fields=tuple(rfs),
                id=ctx.cs.repr_id,
                terse=ctx.cs.terse_repr,
            ),
            orm,
        )

    def generate(self, pl: ReprPlan) -> ta.Iterable[Op]:
        ors: set[Ref] = {REPRLIB_RECURSIVE_REPR_GLOBAL}

        part_lines: list[str] = []

        for f in pl.fields:
            pfx = ''
            if not (pl.terse and not f.kw_only):
                pfx = f'{f.name}='

            if f.fn is not None:
                ors.add(f.fn)
                part_lines.extend([
                    f'    if (s := {f.fn.ident()}(self.{f.name})) is not None:',
                    f'        parts.append(f"{pfx}{{s}}")',
                ])
            else:
                part_lines.append(
                    f'    parts.append(f"{pfx}{{self.{f.name}!r}}")',
                )

        return [
            AddMethodOp(
                '__repr__',
                '\n'.join([
                    f'@{REPRLIB_RECURSIVE_REPR_GLOBAL.ident}()',
                    f'def __repr__(self):',
                    f'    parts = []',
                    *part_lines,
                    f'    return (',
                    f'        f"{{self.__class__.__qualname__}}{'@{hex(id(self))[2:]}' if pl.id else ''}("',
                    f'        f"{{\', \'.join(parts)}}"',
                    f'        f")"',
                    f'    )',
                ]),
                frozenset(ors),
            ),
        ]



"""
TODO:
 - further decomp?
  - Sized + Iterable?
  - COM-style QueryInterface? :|
 - OpKv
 - table ala guava - (row key, column key) keys, sparse storage
 - router
 - value thunker idiom - for key-iterable-only storage, transform Kv[K, V], to Kv[K, ta.Callable[[], V]]
 - zict classes
  - AsyncBuffer
  - Buffer
  - Cache
  - File
  - Func
  - LMDB
  - LRU
  - Sieve
  - Zip
"""
import abc
import typing as ta

from ... import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class Kv(lang.Abstract, ta.Generic[K, V]):
    @abc.abstractmethod
    def __getitem__(self, k: K, /) -> V:
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def items(self) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError


class MutableKv(Kv[K, V], lang.Abstract):  # noqa
    @abc.abstractmethod
    def __setitem__(self, k: K, v: V, /) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, k: K, /) -> None:
        raise NotImplementedError



import collections.abc
import typing as ta

from .iter import fields_dict


##


def _only_test(v: ta.Any) -> bool:
    if v is None:
        return False
    elif isinstance(v, collections.abc.Iterable):
        return bool(v)
    else:
        return True


def only(
        obj: ta.Any,
        *flds: str,
        all: bool = False,  # noqa
        test: ta.Callable[[ta.Any], bool] = _only_test,
) -> bool:
    fdct = fields_dict(obj)
    for fn in flds:
        if fn not in fdct:
            raise KeyError(fn)

    rem = set(flds)
    for fn, f in fdct.items():
        if not f.compare and fn not in rem:
            continue

        v = getattr(obj, fn)
        if test(v):
            if fn in rem:
                rem.remove(fn)
            else:
                return False

    if rem and all:
        return False

    return True



# ruff: noqa: UP006 UP007
# @omlish-lite
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Taneli Hukkinen
# Licensed to PSF under a Contributor Agreement.
#
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022, 2023 Python Software Foundation; All
# Rights Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
#
# https://github.com/python/cpython/blob/9ce90206b7a4649600218cf0bd4826db79c9a312/Lib/tomllib/_parser.py
import datetime
import functools
import re
import string
import types
import typing as ta


TomlParseFloat = ta.Callable[[str], ta.Any]
TomlKey = ta.Tuple[str, ...]
TomlPos = int  # ta.TypeAlias


##


@functools.lru_cache()  # noqa
def toml_cached_tz(hour_str: str, minute_str: str, sign_str: str) -> datetime.timezone:
    sign = 1 if sign_str == '+' else -1
    return datetime.timezone(
        datetime.timedelta(
            hours=sign * int(hour_str),
            minutes=sign * int(minute_str),
        ),
    )


def toml_make_safe_parse_float(parse_float: TomlParseFloat) -> TomlParseFloat:
    """
    A decorator to make `parse_float` safe.

    `parse_float` must not return dicts or lists, because these types would be mixed with parsed TOML tables and arrays,
    thus confusing the parser. The returned decorated callable raises `ValueError` instead of returning illegal types.
    """

    # The default `float` callable never returns illegal types. Optimize it.
    if parse_float is float:
        return float

    def safe_parse_float(float_str: str) -> ta.Any:
        float_value = parse_float(float_str)
        if isinstance(float_value, (dict, list)):
            raise ValueError('parse_float must not return dicts or lists')  # noqa
        return float_value

    return safe_parse_float


class TomlDecodeError(ValueError):
    """An error raised if a document is not valid TOML."""


def toml_load(fp: ta.BinaryIO, /, *, parse_float: TomlParseFloat = float) -> ta.Dict[str, ta.Any]:
    """Parse TOML from a binary file object."""

    b = fp.read()
    try:
        s = b.decode()
    except AttributeError:
        raise TypeError("File must be opened in binary mode, e.g. use `open('foo.toml', 'rb')`") from None
    return toml_loads(s, parse_float=parse_float)


def toml_loads(s: str, /, *, parse_float: TomlParseFloat = float) -> ta.Dict[str, ta.Any]:  # noqa: C901
    """Parse TOML from a string."""

    # The spec allows converting "\r\n" to "\n", even in string literals. Let's do so to simplify parsing.
    try:
        src = s.replace('\r\n', '\n')
    except (AttributeError, TypeError):
        raise TypeError(f"Expected str object, not '{type(s).__qualname__}'") from None

    parse_float = toml_make_safe_parse_float(parse_float)

    parser = TomlParser(
        src,
        parse_float=parse_float,
    )

    return parser.parse()


class TomlFlags:
    """Flags that map to parsed keys/namespaces."""

    # Marks an immutable namespace (inline array or inline table).
    FROZEN = 0
    # Marks a nest that has been explicitly created and can no longer be opened using the "[table]" syntax.
    EXPLICIT_NEST = 1

    def __init__(self) -> None:
        super().__init__()

        self._flags: ta.Dict[str, dict] = {}
        self._pending_flags: ta.Set[ta.Tuple[TomlKey, int]] = set()

    def add_pending(self, key: TomlKey, flag: int) -> None:
        self._pending_flags.add((key, flag))

    def finalize_pending(self) -> None:
        for key, flag in self._pending_flags:
            self.set(key, flag, recursive=False)
        self._pending_flags.clear()

    def unset_all(self, key: TomlKey) -> None:
        cont = self._flags
        for k in key[:-1]:
            if k not in cont:
                return
            cont = cont[k]['nested']
        cont.pop(key[-1], None)

    def set(self, key: TomlKey, flag: int, *, recursive: bool) -> None:  # noqa: A003
        cont = self._flags
        key_parent, key_stem = key[:-1], key[-1]

        for k in key_parent:
            if k not in cont:
                cont[k] = {'flags': set(), 'recursive_flags': set(), 'nested': {}}
            cont = cont[k]['nested']

        if key_stem not in cont:
            cont[key_stem] = {'flags': set(), 'recursive_flags': set(), 'nested': {}}

        cont[key_stem]['recursive_flags' if recursive else 'flags'].add(flag)

    def is_(self, key: TomlKey, flag: int) -> bool:
        if not key:
            return False  # document root has no flags

        cont = self._flags
        for k in key[:-1]:
            if k not in cont:
                return False

            inner_cont = cont[k]
            if flag in inner_cont['recursive_flags']:
                return True

            cont = inner_cont['nested']

        key_stem = key[-1]
        if key_stem in cont:
            cont = cont[key_stem]
            return flag in cont['flags'] or flag in cont['recursive_flags']

        return False


class TomlNestedDict:
    def __init__(self) -> None:
        super().__init__()

        # The parsed content of the TOML document
        self.dict: ta.Dict[str, ta.Any] = {}

    def get_or_create_nest(
            self,
            key: TomlKey,
            *,
            access_lists: bool = True,
    ) -> dict:
        cont: ta.Any = self.dict

        for k in key:
            if k not in cont:
                cont[k] = {}

            cont = cont[k]

            if access_lists and isinstance(cont, list):
                cont = cont[-1]

            if not isinstance(cont, dict):
                raise KeyError('There is no nest behind this key')

        return cont

    def append_nest_to_list(self, key: TomlKey) -> None:
        cont = self.get_or_create_nest(key[:-1])

        last_key = key[-1]
        if last_key in cont:
            list_ = cont[last_key]
            if not isinstance(list_, list):
                raise KeyError('An object other than list found behind this key')
            list_.append({})

        else:
            cont[last_key] = [{}]


class TomlParser:
    def __init__(
            self,
            src: str,
            *,
            parse_float: TomlParseFloat = float,
    ) -> None:
        super().__init__()

        self.src = src

        self.parse_float = parse_float

        self.data = TomlNestedDict()
        self.flags = TomlFlags()
        self.pos = 0

    ASCII_CTRL = frozenset(chr(i) for i in range(32)) | frozenset(chr(127))

    # Neither of these sets include quotation mark or backslash. They are currently handled as separate cases in the
    # parser functions.
    ILLEGAL_BASIC_STR_CHARS = ASCII_CTRL - frozenset('\t')
    ILLEGAL_MULTILINE_BASIC_STR_CHARS = ASCII_CTRL - frozenset('\t\n')

    ILLEGAL_LITERAL_STR_CHARS = ILLEGAL_BASIC_STR_CHARS
    ILLEGAL_MULTILINE_LITERAL_STR_CHARS = ILLEGAL_MULTILINE_BASIC_STR_CHARS

    ILLEGAL_COMMENT_CHARS = ILLEGAL_BASIC_STR_CHARS

    WS = frozenset(' \t')
    WS_AND_NEWLINE = WS | frozenset('\n')
    BARE_KEY_CHARS = frozenset(string.ascii_letters + string.digits + '-_')
    KEY_INITIAL_CHARS = BARE_KEY_CHARS | frozenset("\"'")
    HEXDIGIT_CHARS = frozenset(string.hexdigits)

    BASIC_STR_ESCAPE_REPLACEMENTS = types.MappingProxyType({
        '\\b': '\u0008',  # backspace
        '\\t': '\u0009',  # tab
        '\\n': '\u000A',  # linefeed
        '\\f': '\u000C',  # form feed
        '\\r': '\u000D',  # carriage return
        '\\"': '\u0022',  # quote
        '\\\\': '\u005C',  # backslash
    })

    def parse(self) -> ta.Dict[str, ta.Any]:  # noqa: C901
        header: TomlKey = ()

        # Parse one statement at a time (typically means one line in TOML source)
        while True:
            # 1. Skip line leading whitespace
            self.skip_chars(self.WS)

            # 2. Parse rules. Expect one of the following:
            #    - end of file
            #    - end of line
            #    - comment
            #    - key/value pair
            #    - append dict to list (and move to its namespace)
            #    - create dict (and move to its namespace)
            # Skip trailing whitespace when applicable.
            try:
                char = self.src[self.pos]
            except IndexError:
                break

            if char == '\n':
                self.pos += 1
                continue

            if char in self.KEY_INITIAL_CHARS:
                self.key_value_rule(header)
                self.skip_chars(self.WS)

            elif char == '[':
                try:
                    second_char: ta.Optional[str] = self.src[self.pos + 1]
                except IndexError:
                    second_char = None

                self.flags.finalize_pending()

                if second_char == '[':
                    header = self.create_list_rule()
                else:
                    header = self.create_dict_rule()

                self.skip_chars(self.WS)

            elif char != '#':
                raise self.suffixed_err('Invalid statement')

            # 3. Skip comment
            self.skip_comment()

            # 4. Expect end of line or end of file
            try:
                char = self.src[self.pos]
            except IndexError:
                break

            if char != '\n':
                raise self.suffixed_err('Expected newline or end of document after a statement')

            self.pos += 1

        return self.data.dict

    def skip_chars(self, chars: ta.Iterable[str]) -> None:
        try:
            while self.src[self.pos] in chars:
                self.pos += 1
        except IndexError:
            pass

    def skip_until(
            self,
            expect: str,
            *,
            error_on: ta.FrozenSet[str],
            error_on_eof: bool,
    ) -> None:
        try:
            new_pos = self.src.index(expect, self.pos)
        except ValueError:
            new_pos = len(self.src)
            if error_on_eof:
                raise self.suffixed_err(f'Expected {expect!r}', pos=new_pos) from None

        if not error_on.isdisjoint(self.src[self.pos:new_pos]):
            while self.src[self.pos] not in error_on:
                self.pos += 1

            raise self.suffixed_err(f'Found invalid character {self.src[self.pos]!r}')

        self.pos = new_pos

    def skip_comment(self) -> None:
        try:
            char: ta.Optional[str] = self.src[self.pos]
        except IndexError:
            char = None

        if char == '#':
            self.pos += 1
            self.skip_until(
                '\n',
                error_on=self.ILLEGAL_COMMENT_CHARS,
                error_on_eof=False,
            )

    def skip_comments_and_array_ws(self) -> None:
        while True:
            pos_before_skip = self.pos
            self.skip_chars(self.WS_AND_NEWLINE)
            self.skip_comment()
            if self.pos == pos_before_skip:
                return

    def create_dict_rule(self) -> TomlKey:
        self.pos += 1  # Skip "["
        self.skip_chars(self.WS)
        key = self.parse_key()

        if self.flags.is_(key, TomlFlags.EXPLICIT_NEST) or self.flags.is_(key, TomlFlags.FROZEN):
            raise self.suffixed_err(f'Cannot declare {key} twice')

        self.flags.set(key, TomlFlags.EXPLICIT_NEST, recursive=False)

        try:
            self.data.get_or_create_nest(key)
        except KeyError:
            raise self.suffixed_err('Cannot overwrite a value') from None

        if not self.src.startswith(']', self.pos):
            raise self.suffixed_err("Expected ']' at the end of a table declaration")

        self.pos += 1
        return key

    def create_list_rule(self) -> TomlKey:
        self.pos += 2  # Skip "[["
        self.skip_chars(self.WS)

        key = self.parse_key()

        if self.flags.is_(key, TomlFlags.FROZEN):
            raise self.suffixed_err(f'Cannot mutate immutable namespace {key}')

        # Free the namespace now that it points to another empty list item...
        self.flags.unset_all(key)

        # ...but this key precisely is still prohibited from table declaration
        self.flags.set(key, TomlFlags.EXPLICIT_NEST, recursive=False)

        try:
            self.data.append_nest_to_list(key)
        except KeyError:
            raise self.suffixed_err('Cannot overwrite a value') from None

        if not self.src.startswith(']]', self.pos):
            raise self.suffixed_err("Expected ']]' at the end of an array declaration")

        self.pos += 2
        return key

    def key_value_rule(self, header: TomlKey) -> None:
        key, value = self.parse_key_value_pair()
        key_parent, key_stem = key[:-1], key[-1]
        abs_key_parent = header + key_parent

        relative_path_cont_keys = (header + key[:i] for i in range(1, len(key)))
        for cont_key in relative_path_cont_keys:
            # Check that dotted key syntax does not redefine an existing table
            if self.flags.is_(cont_key, TomlFlags.EXPLICIT_NEST):
                raise self.suffixed_err(f'Cannot redefine namespace {cont_key}')

            # Containers in the relative path can't be opened with the table syntax or dotted key/value syntax in
            # following table sections.
            self.flags.add_pending(cont_key, TomlFlags.EXPLICIT_NEST)

        if self.flags.is_(abs_key_parent, TomlFlags.FROZEN):
            raise self.suffixed_err(f'Cannot mutate immutable namespace {abs_key_parent}')

        try:
            nest = self.data.get_or_create_nest(abs_key_parent)
        except KeyError:
            raise self.suffixed_err('Cannot overwrite a value') from None

        if key_stem in nest:
            raise self.suffixed_err('Cannot overwrite a value')

        # Mark inline table and array namespaces recursively immutable
        if isinstance(value, (dict, list)):
            self.flags.set(header + key, TomlFlags.FROZEN, recursive=True)

        nest[key_stem] = value

    def parse_key_value_pair(self) -> ta.Tuple[TomlKey, ta.Any]:
        key = self.parse_key()

        try:
            char: ta.Optional[str] = self.src[self.pos]
        except IndexError:
            char = None

        if char != '=':
            raise self.suffixed_err("Expected '=' after a key in a key/value pair")

        self.pos += 1
        self.skip_chars(self.WS)

        value = self.parse_value()
        return key, value

    def parse_key(self) -> TomlKey:
        key_part = self.parse_key_part()
        key: TomlKey = (key_part,)

        self.skip_chars(self.WS)

        while True:
            try:
                char: ta.Optional[str] = self.src[self.pos]
            except IndexError:
                char = None

            if char != '.':
                return key

            self.pos += 1
            self.skip_chars(self.WS)

            key_part = self.parse_key_part()
            key += (key_part,)

            self.skip_chars(self.WS)

    def parse_key_part(self) -> str:
        try:
            char: ta.Optional[str] = self.src[self.pos]
        except IndexError:
            char = None

        if char in self.BARE_KEY_CHARS:
            start_pos = self.pos
            self.skip_chars(self.BARE_KEY_CHARS)
            return self.src[start_pos:self.pos]

        if char == "'":
            return self.parse_literal_str()

        if char == '"':
            return self.parse_one_line_basic_str()

        raise self.suffixed_err('Invalid initial character for a key part')

    def parse_one_line_basic_str(self) -> str:
        self.pos += 1
        return self.parse_basic_str(multiline=False)

    def parse_array(self) -> list:
        self.pos += 1
        array: list = []

        self.skip_comments_and_array_ws()
        if self.src.startswith(']', self.pos):
            self.pos += 1
            return array

        while True:
            val = self.parse_value()
            array.append(val)
            self.skip_comments_and_array_ws()

            c = self.src[self.pos:self.pos + 1]
            if c == ']':
                self.pos += 1
                return array

            if c != ',':
                raise self.suffixed_err('Unclosed array')

            self.pos += 1

            self.skip_comments_and_array_ws()

            if self.src.startswith(']', self.pos):
                self.pos += 1
                return array

    def parse_inline_table(self) -> dict:
        self.pos += 1
        nested_dict = TomlNestedDict()
        flags = TomlFlags()

        self.skip_chars(self.WS)

        if self.src.startswith('}', self.pos):
            self.pos += 1
            return nested_dict.dict

        while True:
            key, value = self.parse_key_value_pair()
            key_parent, key_stem = key[:-1], key[-1]

            if flags.is_(key, TomlFlags.FROZEN):
                raise self.suffixed_err(f'Cannot mutate immutable namespace {key}')

            try:
                nest = nested_dict.get_or_create_nest(key_parent, access_lists=False)
            except KeyError:
                raise self.suffixed_err('Cannot overwrite a value') from None

            if key_stem in nest:
                raise self.suffixed_err(f'Duplicate inline table key {key_stem!r}')

            nest[key_stem] = value
            self.skip_chars(self.WS)

            c = self.src[self.pos:self.pos + 1]
            if c == '}':
                self.pos += 1
                return nested_dict.dict

            if c != ',':
                raise self.suffixed_err('Unclosed inline table')

            if isinstance(value, (dict, list)):
                flags.set(key, TomlFlags.FROZEN, recursive=True)

            self.pos += 1
            self.skip_chars(self.WS)

    def parse_basic_str_escape(self, multiline: bool = False) -> str:
        escape_id = self.src[self.pos:self.pos + 2]
        self.pos += 2

        if multiline and escape_id in {'\\ ', '\\\t', '\\\n'}:
            # Skip whitespace until next non-whitespace character or end of the doc. Error if non-whitespace is found
            # before newline.
            if escape_id != '\\\n':
                self.skip_chars(self.WS)

                try:
                    char = self.src[self.pos]
                except IndexError:
                    return ''

                if char != '\n':
                    raise self.suffixed_err("Unescaped '\\' in a string")

                self.pos += 1

            self.skip_chars(self.WS_AND_NEWLINE)
            return ''

        if escape_id == '\\u':
            return self.parse_hex_char(4)

        if escape_id == '\\U':
            return self.parse_hex_char(8)

        try:
            return self.BASIC_STR_ESCAPE_REPLACEMENTS[escape_id]
        except KeyError:
            raise self.suffixed_err("Unescaped '\\' in a string") from None

    def parse_basic_str_escape_multiline(self) -> str:
        return self.parse_basic_str_escape(multiline=True)

    @classmethod
    def is_unicode_scalar_value(cls, codepoint: int) -> bool:
        return (0 <= codepoint <= 55295) or (57344 <= codepoint <= 1114111)

    def parse_hex_char(self, hex_len: int) -> str:
        hex_str = self.src[self.pos:self.pos + hex_len]

        if len(hex_str) != hex_len or not self.HEXDIGIT_CHARS.issuperset(hex_str):
            raise self.suffixed_err('Invalid hex value')

        self.pos += hex_len
        hex_int = int(hex_str, 16)

        if not self.is_unicode_scalar_value(hex_int):
            raise self.suffixed_err('Escaped character is not a Unicode scalar value')

        return chr(hex_int)

    def parse_literal_str(self) -> str:
        self.pos += 1  # Skip starting apostrophe
        start_pos = self.pos
        self.skip_until("'", error_on=self.ILLEGAL_LITERAL_STR_CHARS, error_on_eof=True)
        end_pos = self.pos
        self.pos += 1
        return self.src[start_pos:end_pos]  # Skip ending apostrophe

    def parse_multiline_str(self, *, literal: bool) -> str:
        self.pos += 3
        if self.src.startswith('\n', self.pos):
            self.pos += 1

        if literal:
            delim = "'"
            start_pos = self.pos
            self.skip_until(
                "'''",
                error_on=self.ILLEGAL_MULTILINE_LITERAL_STR_CHARS,
                error_on_eof=True,
            )
            result = self.src[start_pos:self.pos]
            self.pos += 3

        else:
            delim = '"'
            result = self.parse_basic_str(multiline=True)

        # Add at maximum two extra apostrophes/quotes if the end sequence is 4 or 5 chars long instead of just 3.
        if not self.src.startswith(delim, self.pos):
            return result

        self.pos += 1
        if not self.src.startswith(delim, self.pos):
            return result + delim

        self.pos += 1
        return result + (delim * 2)

    def parse_basic_str(self, *, multiline: bool) -> str:
        if multiline:
            error_on = self.ILLEGAL_MULTILINE_BASIC_STR_CHARS
            parse_escapes = self.parse_basic_str_escape_multiline
        else:
            error_on = self.ILLEGAL_BASIC_STR_CHARS
            parse_escapes = self.parse_basic_str_escape

        result = ''
        start_pos = self.pos
        while True:
            try:
                char = self.src[self.pos]
            except IndexError:
                raise self.suffixed_err('Unterminated string') from None

            if char == '"':
                if not multiline:
                    end_pos = self.pos
                    self.pos += 1
                    return result + self.src[start_pos:end_pos]

                if self.src.startswith('"""', self.pos):
                    end_pos = self.pos
                    self.pos += 3
                    return result + self.src[start_pos:end_pos]

                self.pos += 1
                continue

            if char == '\\':
                result += self.src[start_pos:self.pos]
                parsed_escape = parse_escapes()
                result += parsed_escape
                start_pos = self.pos
                continue

            if char in error_on:
                raise self.suffixed_err(f'Illegal character {char!r}')

            self.pos += 1

    def parse_value(self) -> ta.Any:  # noqa: C901
        try:
            char: ta.Optional[str] = self.src[self.pos]
        except IndexError:
            char = None

        # IMPORTANT: order conditions based on speed of checking and likelihood

        # Basic strings
        if char == '"':
            if self.src.startswith('"""', self.pos):
                return self.parse_multiline_str(literal=False)
            return self.parse_one_line_basic_str()

        # Literal strings
        if char == "'":
            if self.src.startswith("'''", self.pos):
                return self.parse_multiline_str(literal=True)
            return self.parse_literal_str()

        # Booleans
        if char == 't':
            if self.src.startswith('true', self.pos):
                self.pos += 4
                return True

        if char == 'f':
            if self.src.startswith('false', self.pos):
                self.pos += 5
                return False

        # Arrays
        if char == '[':
            return self.parse_array()

        # Inline tables
        if char == '{':
            return self.parse_inline_table()

        # Dates and times
        datetime_match = self.RE_DATETIME.match(self.src, self.pos)
        if datetime_match:
            try:
                datetime_obj = self.match_to_datetime(datetime_match)
            except ValueError as e:
                raise self.suffixed_err('Invalid date or datetime') from e

            self.pos = datetime_match.end()
            return datetime_obj

        localtime_match = self.RE_LOCALTIME.match(self.src, self.pos)
        if localtime_match:
            self.pos = localtime_match.end()
            return self.match_to_localtime(localtime_match)

        # Integers and "normal" floats. The regex will greedily match any type starting with a decimal char, so needs to
        # be located after handling of dates and times.
        number_match = self.RE_NUMBER.match(self.src, self.pos)
        if number_match:
            self.pos = number_match.end()
            return self.match_to_number(number_match, self.parse_float)

        # Special floats
        first_three = self.src[self.pos:self.pos + 3]
        if first_three in {'inf', 'nan'}:
            self.pos += 3
            return self.parse_float(first_three)

        first_four = self.src[self.pos:self.pos + 4]
        if first_four in {'-inf', '+inf', '-nan', '+nan'}:
            self.pos += 4
            return self.parse_float(first_four)

        raise self.suffixed_err('Invalid value')

    def coord_repr(self, pos: TomlPos) -> str:
        if pos >= len(self.src):
            return 'end of document'

        line = self.src.count('\n', 0, pos) + 1
        if line == 1:
            column = pos + 1
        else:
            column = pos - self.src.rindex('\n', 0, pos)

        return f'line {line}, column {column}'

    def suffixed_err(self, msg: str, *, pos: ta.Optional[TomlPos] = None) -> TomlDecodeError:
        """Return a `TomlDecodeError` where error message is suffixed with coordinates in source."""

        if pos is None:
            pos = self.pos
        return TomlDecodeError(f'{msg} (at {self.coord_repr(pos)})')

    _TIME_RE_STR = r'([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9])(?:\.([0-9]{1,6})[0-9]*)?'

    RE_NUMBER = re.compile(
        r"""
        0
        (?:
            x[0-9A-Fa-f](?:_?[0-9A-Fa-f])*   # hex
            |
            b[01](?:_?[01])*                 # bin
            |
            o[0-7](?:_?[0-7])*               # oct
        )
        |
        [+-]?(?:0|[1-9](?:_?[0-9])*)         # dec, integer part
        (?P<floatpart>
            (?:\.[0-9](?:_?[0-9])*)?         # optional fractional part
            (?:[eE][+-]?[0-9](?:_?[0-9])*)?  # optional exponent part
        )
        """,
        flags=re.VERBOSE,
    )

    RE_LOCALTIME = re.compile(_TIME_RE_STR)

    RE_DATETIME = re.compile(
        rf"""
        ([0-9]{{4}})-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])  # date, e.g. 1988-10-27
        (?:
            [Tt ]
            {_TIME_RE_STR}
            (?:([Zz])|([+-])([01][0-9]|2[0-3]):([0-5][0-9]))?  # optional time offset
        )?
        """,
        flags=re.VERBOSE,
    )

    @classmethod
    def match_to_datetime(cls, match: re.Match) -> ta.Union[datetime.datetime, datetime.date]:
        """
        Convert a `RE_DATETIME` match to `datetime.datetime` or `datetime.date`.

        Raises ValueError if the match does not correspond to a valid date or datetime.
        """

        (
            year_str,
            month_str,
            day_str,
            hour_str,
            minute_str,
            sec_str,
            micros_str,
            zulu_time,
            offset_sign_str,
            offset_hour_str,
            offset_minute_str,
        ) = match.groups()

        year, month, day = int(year_str), int(month_str), int(day_str)

        if hour_str is None:
            return datetime.date(year, month, day)

        hour, minute, sec = int(hour_str), int(minute_str), int(sec_str)

        micros = int(micros_str.ljust(6, '0')) if micros_str else 0

        if offset_sign_str:
            tz: ta.Optional[datetime.tzinfo] = toml_cached_tz(
                offset_hour_str, offset_minute_str, offset_sign_str,
            )
        elif zulu_time:
            tz = datetime.UTC
        else:  # local date-time
            tz = None

        return datetime.datetime(year, month, day, hour, minute, sec, micros, tzinfo=tz)

    @classmethod
    def match_to_localtime(cls, match: re.Match) -> datetime.time:
        hour_str, minute_str, sec_str, micros_str = match.groups()
        micros = int(micros_str.ljust(6, '0')) if micros_str else 0
        return datetime.time(int(hour_str), int(minute_str), int(sec_str), micros)

    @classmethod
    def match_to_number(cls, match: re.Match, parse_float: TomlParseFloat) -> ta.Any:
        if match.group('floatpart'):
            return parse_float(match.group())
        return int(match.group(), 0)



import dataclasses as dc
import functools
import itertools
import typing as ta


T = ta.TypeVar('T')
R = ta.TypeVar('R')


##


def ilen(it: ta.Iterable) -> int:
    c = 0
    for _ in it:
        c += 1
    return c


def take(n: int, it: ta.Iterable[T]) -> list[T]:
    return list(itertools.islice(it, n))


def exhaust(it: ta.Iterable[ta.Any]) -> None:
    for _ in it:
        pass


def peek(vs: ta.Iterable[T]) -> tuple[T, ta.Iterator[T]]:
    it = iter(vs)
    v = next(it)
    return v, itertools.chain(iter((v,)), it)


def interleave(vs: ta.Iterable[T], d: T) -> ta.Iterable[T]:
    for i, v in enumerate(vs):
        if i:
            yield d
        yield v


def renumerate(it: ta.Iterable[T]) -> ta.Iterable[tuple[T, int]]:
    return ((e, i) for i, e in enumerate(it))


def common_prefix_len(*its: ta.Iterable) -> int:
    return ilen(itertools.takewhile(lambda t: all(e == t[0] for e in t[1:]), zip(*its)))


##


@ta.overload
def readiter(f: ta.TextIO, sz: int) -> ta.Iterator[str]:
    ...


@ta.overload
def readiter(f: ta.BinaryIO, sz: int) -> ta.Iterator[bytes]:
    ...


@ta.overload
def readiter(f: ta.IO, sz: int) -> ta.Iterator[ta.AnyStr]:
    ...


def readiter(f, sz):
    return iter(functools.partial(f.read, sz), None)


##


@dc.dataclass(frozen=True)
class IterGen(ta.Generic[T]):
    fn: ta.Callable[[], ta.Iterable[T]]

    def __iter__(self):
        return iter(self.fn())


itergen = IterGen


##


flatten = itertools.chain.from_iterable


def flatmap(fn: ta.Callable[[T], ta.Iterable[R]], it: ta.Iterable[T]) -> ta.Iterable[R]:
    return flatten(map(fn, it))


##


Rangeable: ta.TypeAlias = ta.Union[  # noqa
    int,
    tuple[int],
    tuple[int, int],
    tuple[int, int, int],
    ta.Iterable[int],
]


def asrange(i: Rangeable) -> ta.Iterable[int]:
    if isinstance(i, int):
        return range(i)
    elif isinstance(i, tuple):
        return range(*i)
    elif isinstance(i, ta.Iterable):
        return i
    else:
        raise TypeError(i)


def prodrange(*dims: Rangeable) -> ta.Iterable[ta.Sequence[int]]:
    if not dims:
        return []
    return itertools.product(*map(asrange, dims))



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

from io import StringIO
from .Token import Token

from .CommonTokenStream import CommonTokenStream


class TokenStreamRewriter(object):
    __slots__ = ('tokens', 'programs', 'lastRewriteTokenIndexes')

    DEFAULT_PROGRAM_NAME = "default"
    PROGRAM_INIT_SIZE = 100
    MIN_TOKEN_INDEX = 0

    def __init__(self, tokens):
        """
        :type  tokens: antlr4.BufferedTokenStream.BufferedTokenStream
        :param tokens:
        :return:
        """
        super(TokenStreamRewriter, self).__init__()
        self.tokens = tokens
        self.programs = {self.DEFAULT_PROGRAM_NAME: []}
        self.lastRewriteTokenIndexes = {}

    def getTokenStream(self):
        return self.tokens

    def rollback(self, instruction_index, program_name):
        ins = self.programs.get(program_name, None)
        if ins:
            self.programs[program_name] = ins[self.MIN_TOKEN_INDEX: instruction_index]

    def deleteProgram(self, program_name=DEFAULT_PROGRAM_NAME):
        self.rollback(self.MIN_TOKEN_INDEX, program_name)

    def insertAfterToken(self, token, text, program_name=DEFAULT_PROGRAM_NAME):
        self.insertAfter(token.tokenIndex, text, program_name)

    def insertAfter(self, index, text, program_name=DEFAULT_PROGRAM_NAME):
        op = self.InsertAfterOp(self.tokens, index + 1, text)
        rewrites = self.getProgram(program_name)
        op.instructionIndex = len(rewrites)
        rewrites.append(op)

    def insertBeforeIndex(self, index, text):
        self.insertBefore(self.DEFAULT_PROGRAM_NAME, index, text)

    def insertBeforeToken(self, token, text, program_name=DEFAULT_PROGRAM_NAME):
        self.insertBefore(program_name, token.tokenIndex, text)

    def insertBefore(self, program_name, index, text):
        op = self.InsertBeforeOp(self.tokens, index, text)
        rewrites = self.getProgram(program_name)
        op.instructionIndex = len(rewrites)
        rewrites.append(op)

    def replaceIndex(self, index, text):
        self.replace(self.DEFAULT_PROGRAM_NAME, index, index, text)

    def replaceRange(self, from_idx, to_idx, text):
        self.replace(self.DEFAULT_PROGRAM_NAME, from_idx, to_idx, text)

    def replaceSingleToken(self, token, text):
        self.replace(self.DEFAULT_PROGRAM_NAME, token.tokenIndex, token.tokenIndex, text)

    def replaceRangeTokens(self, from_token, to_token, text, program_name=DEFAULT_PROGRAM_NAME):
        self.replace(program_name, from_token.tokenIndex, to_token.tokenIndex, text)

    def replace(self, program_name, from_idx, to_idx, text):
        if any((from_idx > to_idx, from_idx < 0, to_idx < 0, to_idx >= len(self.tokens.tokens))):
            raise ValueError(
                'replace: range invalid: {}..{}(size={})'.format(from_idx, to_idx, len(self.tokens.tokens)))
        op = self.ReplaceOp(from_idx, to_idx, self.tokens, text)
        rewrites = self.getProgram(program_name)
        op.instructionIndex = len(rewrites)
        rewrites.append(op)

    def deleteToken(self, token):
        self.delete(self.DEFAULT_PROGRAM_NAME, token, token)

    def deleteIndex(self, index):
        self.delete(self.DEFAULT_PROGRAM_NAME, index, index)

    def delete(self, program_name, from_idx, to_idx):
        if isinstance(from_idx, Token):
            self.replace(program_name, from_idx.tokenIndex, to_idx.tokenIndex, "")
        else:
            self.replace(program_name, from_idx, to_idx, "")

    def lastRewriteTokenIndex(self, program_name=DEFAULT_PROGRAM_NAME):
        return self.lastRewriteTokenIndexes.get(program_name, -1)

    def setLastRewriteTokenIndex(self, program_name, i):
        self.lastRewriteTokenIndexes[program_name] = i

    def getProgram(self, program_name):
        return self.programs.setdefault(program_name, [])

    def getDefaultText(self):
        return self.getText(self.DEFAULT_PROGRAM_NAME, 0, len(self.tokens.tokens) - 1)

    def getText(self, program_name, start:int, stop:int):
        """
        :return: the text in tokens[start, stop](closed interval)
        """
        rewrites = self.programs.get(program_name)

        # ensure start/end are in range
        if stop > len(self.tokens.tokens) - 1:
            stop = len(self.tokens.tokens) - 1
        if start < 0:
            start = 0

        # if no instructions to execute
        if not rewrites: return self.tokens.getText(start, stop)
        buf = StringIO()
        indexToOp = self._reduceToSingleOperationPerIndex(rewrites)
        i = start
        while all((i <= stop, i < len(self.tokens.tokens))):
            op = indexToOp.pop(i, None)
            token = self.tokens.get(i)
            if op is None:
                if token.type != Token.EOF: buf.write(token.text)
                i += 1
            else:
                i = op.execute(buf)

        if stop == len(self.tokens.tokens)-1:
            for op in indexToOp.values():
                if op.index >= len(self.tokens.tokens)-1: buf.write(op.text)

        return buf.getvalue()

    def _reduceToSingleOperationPerIndex(self, rewrites):
        # Walk replaces
        for i, rop in enumerate(rewrites):
            if any((rop is None, not isinstance(rop, TokenStreamRewriter.ReplaceOp))):
                continue
            # Wipe prior inserts within range
            inserts = [op for op in rewrites[:i] if isinstance(op, TokenStreamRewriter.InsertBeforeOp)]
            for iop in inserts:
                if iop.index == rop.index:
                    rewrites[iop.instructionIndex] = None
                    rop.text = '{}{}'.format(iop.text, rop.text)
                elif all((iop.index > rop.index, iop.index <= rop.last_index)):
                    rewrites[iop.instructionIndex] = None

            # Drop any prior replaces contained within
            prevReplaces = [op for op in rewrites[:i] if isinstance(op, TokenStreamRewriter.ReplaceOp)]
            for prevRop in prevReplaces:
                if all((prevRop.index >= rop.index, prevRop.last_index <= rop.last_index)):
                    rewrites[prevRop.instructionIndex] = None
                    continue
                isDisjoint = any((prevRop.last_index<rop.index, prevRop.index>rop.last_index))
                if all((prevRop.text is None, rop.text is None, not isDisjoint)):
                    rewrites[prevRop.instructionIndex] = None
                    rop.index = min(prevRop.index, rop.index)
                    rop.last_index = min(prevRop.last_index, rop.last_index)
                    print('New rop {}'.format(rop))
                elif (not(isDisjoint)):
                    raise ValueError("replace op boundaries of {} overlap with previous {}".format(rop, prevRop))

        # Walk inserts
        for i, iop in enumerate(rewrites):
            if any((iop is None, not isinstance(iop, TokenStreamRewriter.InsertBeforeOp))):
                continue
            prevInserts = [op for op in rewrites[:i] if isinstance(op, TokenStreamRewriter.InsertBeforeOp)]
            for prev_index, prevIop in enumerate(prevInserts):
                if prevIop.index == iop.index and type(prevIop) is TokenStreamRewriter.InsertBeforeOp:
                    iop.text += prevIop.text
                    rewrites[prev_index] = None
                elif prevIop.index == iop.index and type(prevIop) is TokenStreamRewriter.InsertAfterOp:
                    iop.text = prevIop.text + iop.text
                    rewrites[prev_index] = None
            # look for replaces where iop.index is in range; error
            prevReplaces = [op for op in rewrites[:i] if isinstance(op, TokenStreamRewriter.ReplaceOp)]
            for rop in prevReplaces:
                if iop.index == rop.index:
                    rop.text = iop.text + rop.text
                    rewrites[i] = None
                    continue
                if all((iop.index >= rop.index, iop.index <= rop.last_index)):
                    raise ValueError("insert op {} within boundaries of previous {}".format(iop, rop))

        reduced = {}
        for i, op in enumerate(rewrites):
            if op is None: continue
            if reduced.get(op.index): raise ValueError('should be only one op per index')
            reduced[op.index] = op

        return reduced

    class RewriteOperation(object):
        __slots__ = ('tokens', 'index', 'text', 'instructionIndex')

        def __init__(self, tokens, index, text=""):
            """
            :type tokens: CommonTokenStream
            :param tokens:
            :param index:
            :param text:
            :return:
            """
            self.tokens = tokens
            self.index = index
            self.text = text
            self.instructionIndex = 0

        def execute(self, buf):
            """
            :type buf: StringIO.StringIO
            :param buf:
            :return:
            """
            return self.index

        def __str__(self):
            return '<{}@{}:"{}">'.format(self.__class__.__name__, self.tokens.get(self.index), self.text)

    class InsertBeforeOp(RewriteOperation):

        def __init__(self, tokens, index, text=""):
            super(TokenStreamRewriter.InsertBeforeOp, self).__init__(tokens, index, text)

        def execute(self, buf):
            buf.write(self.text)
            if self.tokens.get(self.index).type != Token.EOF:
                buf.write(self.tokens.get(self.index).text)
            return self.index + 1

    class InsertAfterOp(InsertBeforeOp):
        pass

    class ReplaceOp(RewriteOperation):
        __slots__ = 'last_index'

        def __init__(self, from_idx, to_idx, tokens, text):
            super(TokenStreamRewriter.ReplaceOp, self).__init__(tokens, from_idx, text)
            self.last_index = to_idx

        def execute(self, buf):
            if self.text:
                buf.write(self.text)
            return self.last_index + 1

        def __str__(self):
            if self.text:
                return '<ReplaceOp@{}..{}:"{}">'.format(self.tokens.get(self.index), self.tokens.get(self.last_index),
                                                        self.text)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - docstring
 - timebomb
 - auto-discover available ports
"""
import dataclasses as dc
import os
import typing as ta


##


@dc.dataclass(frozen=True)
class DockerPortRelay:
    """
    Uses roughly the following command to forward connections from inside docker-for-mac's vm to the mac host:

      docker run --rm -i -p 5001:5000 alpine/socat -d -d TCP-LISTEN:5000,fork,reuseaddr TCP:host.docker.internal:5021

    This allows requests made by the docker daemon running inside the vm to `host.docker.internal:5001` to be forwarded
    to the mac host on port 5021. The reason for this is to be able to use a docker registry running locally directly on
    the host mac - specifically to be able to do so with ssl certificate checking disabled (which docker will only do on
    localhost, which on a mac in the vm isn't actually the mac host - hence the necessity of the relay).
    """

    docker_port: int  # port
    host_port: int

    name: ta.Optional[str] = None

    DEFAULT_HOST_NAME: ta.ClassVar[str] = 'host.docker.internal'
    host_name: str = DEFAULT_HOST_NAME

    DEFAULT_INTERMEDIATE_PORT: ta.ClassVar[int] = 5000
    intermediate_port: int = DEFAULT_INTERMEDIATE_PORT

    DEFAULT_IMAGE: ta.ClassVar[str] = 'alpine/socat'
    image: str = DEFAULT_IMAGE

    def socat_args(self) -> ta.List[str]:
        return [
            '-d',
            f'TCP-LISTEN:{self.intermediate_port},fork,reuseaddr',
            f'TCP:{self.host_name}:{self.host_port}',
        ]

    def run_args(self) -> ta.List[str]:
        if (name := self.name) is None:
            name = f'docker_port_relay-{os.getpid()}-{self.docker_port}-{self.intermediate_port}-{self.host_port}'

        return [
            '--name', name,
            '--rm',
            '-p', f'{self.docker_port}:{self.intermediate_port}',
            self.image,
            *self.socat_args(),
        ]

    def run_cmd(self) -> ta.List[str]:
        return [
            'docker',
            'run',
            '-i',
            *self.run_args(),
        ]



# ruff: noqa: I001
"""
TODO:
 - delimited.py / jsonl
  - + record separators ala https://en.wikipedia.org/wiki/JSON_streaming
"""
import typing as _ta

from ... import lang as _lang


from .consts import (  # noqa
    COMPACT_KWARGS,
    COMPACT_SEPARATORS,
    PRETTY_INDENT,
    PRETTY_KWARGS,
    PRETTY_SEPARATORS,
    Separators,
)

from .encoding import (  # noqa
    decodes,
    detect_encoding,
)

from .json import (  # noqa
    dump,
    dump_compact,
    dump_pretty,
    dumps,
    dumps_compact,
    dumps_pretty,
    load,
    loads,
)

if _ta.TYPE_CHECKING:
    from .rendering import (  # noqa
        JsonRenderer,
    )
else:
    _lang.proxy_init(globals(), '.rendering', [
        'JsonRenderer',
    ])

from .types import (  # noqa
    SCALAR_TYPES,
    Scalar,
)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - translate json keys
"""
import logging
import typing as ta

from ..lite.json import json_dumps_compact


##


class JsonLogFormatter(logging.Formatter):
    KEYS: ta.Mapping[str, bool] = {
        'name': False,
        'msg': False,
        'args': False,
        'levelname': False,
        'levelno': False,
        'pathname': False,
        'filename': False,
        'module': False,
        'exc_info': True,
        'exc_text': True,
        'stack_info': True,
        'lineno': False,
        'funcName': False,
        'created': False,
        'msecs': False,
        'relativeCreated': False,
        'thread': False,
        'threadName': False,
        'processName': False,
        'process': False,
    }

    def __init__(
            self,
            *args: ta.Any,
            json_dumps: ta.Optional[ta.Callable[[ta.Any], str]] = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        if json_dumps is None:
            json_dumps = json_dumps_compact
        self._json_dumps = json_dumps

    def format(self, record: logging.LogRecord) -> str:
        dct = {
            k: v
            for k, o in self.KEYS.items()
            for v in [getattr(record, k)]
            if not (o and v is None)
        }
        return self._json_dumps(dct)



# ruff: noqa: I001
from .classes import (  # noqa
    Configurable,
)

from .formats import (  # noqa
    ConfigData as Data,

    ConfigLoader as Loader,

    ConfigRenderer as Renderer,

    ObjConfigData as ObjData,

    JsonConfigData as JsonData,
    JsonConfigLoader as JsonLoader,
    JsonConfigRenderer as JsonRenderer,

    TomlConfigData as TomlData,
    TomlConfigLoader as TomlLoader,
    TomlConfigRenderer as TomlRenderer,

    YamlConfigData as YamlData,
    YamlConfigLoader as YamlLoader,
    YamlConfigRenderer as YamlRenderer,

    IniConfigData as IniData,
    IniConfigLoader as IniLoader,
    IniConfigRenderer as IniRenderer,

    SwitchedConfigFileLoader as SwitchedFileLoader,

    DEFAULT_CONFIG_LOADERS as DEFAULT_LOADERS,
    DEFAULT_CONFIG_LOADER as DEFAULT_LOADER,
    DEFAULT_CONFIG_FILE_LOADER as DEFAULT_FILE_LOADER,

    SwitchedConfigRenderer as SwitchedRenderer,

    DEFAULT_CONFIG_RENDERERS as DEFAULT_RENDERERS,
    DEFAULT_CONFIG_RENDERER as DEFAULT_RENDERER,
)

from .types import (  # noqa
    ConfigMap as Map,
)

from .processing import all as processing  # noqa



import typing as ta

from ... import dataclasses as dc
from ... import lang
from ..elements import Element
from ..injector import Injector
from ..multis import MapBinding
from ..multis import MapProvider
from ..multis import SetBinding
from ..multis import SetProvider
from ..providers import LinkProvider
from ..providers import Provider
from .providers import LinkProviderImpl
from .providers import ProviderImpl


##


@dc.dataclass(frozen=True, eq=False)
class SetProviderImpl(ProviderImpl, lang.Final):
    ps: ta.Sequence[ProviderImpl]

    @property
    def providers(self) -> ta.Iterable[Provider]:
        for p in self.ps:
            yield from p.providers

    def provide(self, injector: Injector) -> ta.Any:
        rv = set()
        for ep in self.ps:
            o = ep.provide(injector)
            rv.add(o)
        return rv


@dc.dataclass(frozen=True, eq=False)
class MapProviderImpl(ProviderImpl, lang.Final):
    class Entry(ta.NamedTuple):
        k: ta.Any
        v: ProviderImpl

    es: ta.Sequence[Entry]

    @property
    def providers(self) -> ta.Iterable[Provider]:
        for e in self.es:
            yield from e.v.providers

    def provide(self, injector: Injector) -> ta.Any:
        rv = {}
        for e in self.es:
            o = e.v.provide(injector)
            rv[e.k] = o
        return rv


def make_multi_provider_impl(p: Provider, es_by_ty: ta.MutableMapping[type, list[Element]]) -> ProviderImpl:
    if isinstance(p, SetProvider):
        sbs: ta.Iterable[SetBinding] = es_by_ty.pop(SetBinding, ())  # type: ignore
        return SetProviderImpl([
            LinkProviderImpl(LinkProvider(sb.dst))
            for sb in sbs
        ])

    elif isinstance(p, MapProvider):
        mbs: ta.Iterable[MapBinding] = es_by_ty.pop(MapBinding, ())  # type: ignore
        return MapProviderImpl([
            MapProviderImpl.Entry(
                mb.map_key,
                LinkProviderImpl(LinkProvider(mb.dst)),
            )
            for mb in mbs
        ])

    else:
        raise TypeError(p)



import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import lang
from .abstract import AbstractLifecycle
from .base import Lifecycle
from .controller import LifecycleController
from .states import LifecycleState
from .states import LifecycleStateError
from .states import LifecycleStates


class LifecycleManager(AbstractLifecycle):
    @dc.dataclass(frozen=True)
    class Entry(lang.Final):
        controller: LifecycleController
        dependencies: ta.MutableSet['LifecycleManager.Entry'] = dc.field(default_factory=col.IdentitySet)
        dependents: ta.MutableSet['LifecycleManager.Entry'] = dc.field(default_factory=col.IdentitySet)

    def __init__(
            self,
            *,
            lock: lang.DefaultLockable = None,
    ) -> None:
        super().__init__()

        self._lock = lang.default_lock(lock, False)

        self._entries_by_lifecycle: ta.MutableMapping[Lifecycle, LifecycleManager.Entry] = col.IdentityKeyDict()

        self._controller = LifecycleController(self._lifecycle, lock=self._lock)

    @property
    def controller(self) -> LifecycleController:
        return self._controller

    @property
    def state(self) -> LifecycleState:
        return self._controller.state

    @staticmethod
    def _get_controller(lifecycle: Lifecycle) -> LifecycleController:
        if isinstance(lifecycle, LifecycleController):
            return lifecycle
        # elif isinstance(lifecycle, AbstractLifecycle):
        #     return lifecycle.lifecycle_controller
        elif isinstance(lifecycle, Lifecycle):
            return LifecycleController(lifecycle)
        else:
            raise TypeError(lifecycle)

    def _add_internal(self, lifecycle: Lifecycle, dependencies: ta.Iterable[Lifecycle]) -> Entry:
        check.state(self.state < LifecycleStates.STOPPING and not self.state.is_failed)

        check.isinstance(lifecycle, Lifecycle)
        try:
            entry = self._entries_by_lifecycle[lifecycle]
        except KeyError:
            controller = self._get_controller(lifecycle)
            entry = self._entries_by_lifecycle[lifecycle] = LifecycleManager.Entry(controller)

        for dep in dependencies:
            check.isinstance(dep, Lifecycle)
            dep_entry = self._add_internal(dep, [])
            entry.dependencies.add(dep_entry)
            dep_entry.dependents.add(entry)

        return entry

    def add(
            self,
            lifecycle: Lifecycle,
            dependencies: ta.Iterable[Lifecycle] = (),
    ) -> Entry:
        check.state(self.state < LifecycleStates.STOPPING and not self.state.is_failed)

        with self._lock():
            entry = self._add_internal(lifecycle, dependencies)

            if self.state >= LifecycleStates.CONSTRUCTING:
                def rec(e):
                    if e.controller.state < LifecycleStates.CONSTRUCTED:
                        for dep in e.dependencies:
                            rec(dep)
                        e.controller.lifecycle_construct()
                rec(entry)

            if self.state >= LifecycleStates.STARTING:
                def rec(e):
                    if e.controller.state < LifecycleStates.STARTED:
                        for dep in e.dependencies:
                            rec(dep)
                        e.controller.lifecycle_start()
                rec(entry)

            return entry

    ##

    @ta.override
    def _lifecycle_construct(self) -> None:
        def rec(entry: LifecycleManager.Entry) -> None:
            for dep in entry.dependencies:
                rec(dep)

            if entry.controller.state.is_failed:
                raise LifecycleStateError(entry.controller)

            if entry.controller.state < LifecycleStates.CONSTRUCTED:
                entry.controller.lifecycle_construct()

        for entry in self._entries_by_lifecycle.values():
            rec(entry)

    @ta.override
    def _lifecycle_start(self) -> None:
        def rec(entry: LifecycleManager.Entry) -> None:
            for dep in entry.dependencies:
                rec(dep)

            if entry.controller.state.is_failed:
                raise LifecycleStateError(entry.controller)

            if entry.controller.state < LifecycleStates.CONSTRUCTED:
                entry.controller.lifecycle_construct()

            if entry.controller.state < LifecycleStates.STARTED:
                entry.controller.lifecycle_start()

        for entry in self._entries_by_lifecycle.values():
            rec(entry)

    @ta.override
    def _lifecycle_stop(self) -> None:
        def rec(entry: LifecycleManager.Entry) -> None:
            for dep in entry.dependents:
                rec(dep)

            if entry.controller.state.is_failed:
                raise LifecycleStateError(entry.controller)

            if entry.controller.state is LifecycleStates.STARTED:
                entry.controller.lifecycle_stop()

        for entry in self._entries_by_lifecycle.values():
            rec(entry)

    @ta.override
    def _lifecycle_destroy(self) -> None:
        def rec(entry: LifecycleManager.Entry) -> None:
            for dep in entry.dependents:
                rec(dep)

            if entry.controller.state < LifecycleStates.DESTROYED:
                entry.controller.lifecycle_destroy()

        for entry in self._entries_by_lifecycle.values():
            rec(entry)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - smarter merging than just dumb dict-squashing
"""
import typing as ta

from ..types import ConfigMap


##


def merge_configs(*ms: ConfigMap) -> ConfigMap:
    def rec(o, i):
        for k, v in i.items():
            try:
                e = o[k]
            except KeyError:
                o[k] = v
            else:
                if isinstance(e, ta.Mapping) and isinstance(v, ta.Mapping):
                    rec(e, v)  # noqa
                else:
                    if isinstance(v, ta.Mapping):
                        v = dict(v)
                    o[k] = v

    o: dict = {}
    for i in ms:
        rec(o, i)
    return o



import typing as ta

from .. import lang
from .codecs import make_bytes_object_codec
from .codecs import make_object_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import cbor2
else:
    cbor2 = lang.proxy_import('cbor2')


##


def dump(obj: ta.Any) -> bytes:
    return cbor2.dumps(obj)


def load(s: bytes) -> ta.Any:
    return cbor2.loads(s)


##


CBOR_CODEC = make_bytes_object_codec('cbor', dump, load)

# @omlish-manifest
_CBOR_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'CBOR_CODEC', CBOR_CODEC)



import typing as ta

from ..json import Scalar
from ..json.literals import ESCAPE_MAP
from ..json.literals import encode_string
from ..json.literals import encode_string_ascii
from ..json.rendering import JsonRenderer
from ..json.rendering import JsonRendererOut


##


MULTILINE_STRINGS_ESCAPE_MAP = {
    **ESCAPE_MAP,
    '\n': '\\n\\\n',
}


class Json5Renderer(JsonRenderer):
    def __init__(
            self,
            out: JsonRendererOut,
            *,
            multiline_strings: bool = False,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(out, **kwargs)

        self._multiline_strings = multiline_strings

    def _format_scalar(self, o: Scalar) -> str:
        if (
                self._multiline_strings and
                isinstance(o, str) and
                '\n' in o
        ):
            if self._ensure_ascii:
                return encode_string_ascii(o, escape_map=MULTILINE_STRINGS_ESCAPE_MAP)
            else:
                return encode_string(o, escape_map=MULTILINE_STRINGS_ESCAPE_MAP)

        else:
            return super()._format_scalar(o)



from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import marshal as msh
from .. import reflect as rfl
from ..funcs import match as mfs
from .collection import TypedValues
from .reflect import reflect_typed_values_impls
from .values import ScalarTypedValue
from .values import TypedValue


##


def _build_typed_value_poly(rty: rfl.Type) -> msh.Polymorphism:
    ty: type[TypedValue] = check.issubclass(check.isinstance(rty, type), TypedValue)
    check.state(lang.is_abstract_class(ty))
    return msh.polymorphism_from_subclasses(
        ty,
        naming=msh.Naming.SNAKE,
        strip_suffix='auto',
    )


class TypedValueMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, type) and
        issubclass(rty, ScalarTypedValue) and
        not lang.is_abstract_class(rty)
    ))
    def _build_scalar(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        dc_rfl = dc.reflect(check.isinstance(rty, type))
        v_rty = check.single(dc_rfl.fields_inspection.generic_replaced_field_annotations.values())
        v_m = ctx.make(v_rty)
        return msh.WrappedMarshaler(lambda _, o: o.v, v_m)

    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, type) and
        issubclass(rty, TypedValue) and
        lang.is_abstract_class(rty)
    ))
    def _build_abstract(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        return msh.make_polymorphism_marshaler(
            _build_typed_value_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


class TypedValueUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, type) and
        issubclass(rty, ScalarTypedValue) and
        not lang.is_abstract_class(rty)
    ))
    def _build_scalar(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        rty = check.isinstance(rty, type)
        dc_rfl = dc.reflect(rty)
        v_rty = check.single(dc_rfl.fields_inspection.generic_replaced_field_annotations.values())
        v_u = ctx.make(v_rty)
        return msh.WrappedUnmarshaler(lambda _, v: rty(v), v_u)

    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, type) and
        issubclass(rty, TypedValue) and
        lang.is_abstract_class(rty)
    ))
    def _build_abstract(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        return msh.make_polymorphism_unmarshaler(
            _build_typed_value_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


##


def _build_typed_values_impls(rty: rfl.Type) -> msh.Impls:
    gty = check.isinstance(rty, rfl.Generic)
    check.is_(gty.cls, TypedValues)

    tv_cls_set = reflect_typed_values_impls(check.single(gty.args))

    tv_impls: list[msh.Impl] = []
    for tv_cls in tv_cls_set:
        tv_impls.extend(_build_typed_value_poly(tv_cls).impls)

    return msh.Impls(tv_impls)


class TypedValuesMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is TypedValues)
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        tv_m = msh.make_polymorphism_marshaler(
            msh.Impls(_build_typed_values_impls(rty)),
            msh.WrapperTypeTagging(),
            ctx,
        )
        return msh.IterableMarshaler(tv_m)


class TypedValuesUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is TypedValues)
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        tv_u = msh.make_polymorphism_unmarshaler(
            msh.Impls(_build_typed_values_impls(rty)),
            msh.WrapperTypeTagging(),
            ctx,
        )
        return msh.IterableUnmarshaler(lambda it: TypedValues(*it), tv_u)  # noqa


##


@lang.static_init
def _install_standard_marshalling() -> None:
    msh.install_standard_factories(
        TypedValueMarshalerFactory(),
        TypedValueUnmarshalerFactory(),

        TypedValuesMarshalerFactory(),
        TypedValuesUnmarshalerFactory(),
    )



def read_package_resource_binary(package: str, resource: str) -> bytes:
    import importlib.resources
    return importlib.resources.read_binary(package, resource)


def read_package_resource_text(package: str, resource: str) -> str:
    import importlib.resources
    return importlib.resources.read_text(package, resource)



import typing as ta

from ... import lang
from .base import Kv
from .base import MutableKv
from .wrappers import SimpleWrapperKv
from .wrappers import SimpleWrapperMutableKv


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class KeyFilteredKv(SimpleWrapperKv[K, V]):
    def __init__(
            self,
            u: Kv[K, V],
            fn: ta.Callable[[K], bool],
    ) -> None:
        super().__init__(u)

        self._fn = fn

    def __getitem__(self, k: K, /) -> V:
        if not self._fn(k):
            raise KeyError(k)
        return self._u[k]

    def __len__(self) -> int:
        return lang.ilen(self.items())

    def items(self) -> ta.Iterator[tuple[K, V]]:
        fn = self._fn
        return filter(lambda t: fn(t[0]), self._u.items())


class KeyFilteredMutableKv(KeyFilteredKv[K, V], SimpleWrapperMutableKv[K, V]):
    def __init__(
            self,
            u: MutableKv[K, V],
            fn: ta.Callable[[K], bool],
    ) -> None:
        super().__init__(u, fn)

    def __setitem__(self, k: K, v: V, /) -> None:
        if not self._fn(k):
            raise KeyError(k)
        self._u[k] = v

    def __delitem__(self, k: K, /) -> None:
        if not self._fn(k):
            raise KeyError(k)
        del self._u[k]


##


class ValueFilteredKeyError(KeyError):
    pass


class ValueFilteredKv(SimpleWrapperKv[K, V]):
    def __init__(
            self,
            u: Kv[K, V],
            fn: ta.Callable[[V], bool],
    ) -> None:
        super().__init__(u)

        self._fn = fn

    def __getitem__(self, k: K, /) -> V:
        v = self._u[k]
        if not self._fn(v):
            raise ValueFilteredKeyError(k)
        return v

    def __len__(self) -> int:
        return lang.ilen(self.items())

    def items(self) -> ta.Iterator[tuple[K, V]]:
        fn = self._fn
        return ((k, v) for k, v in self._u.items() if fn(v))


class ValueFilteredMutableKv(ValueFilteredKv[K, V], SimpleWrapperMutableKv[K, V]):
    def __init__(
            self,
            u: MutableKv[K, V],
            fn: ta.Callable[[V], bool],
    ) -> None:
        super().__init__(u, fn)

    def __setitem__(self, k: K, v: V, /) -> None:
        if not self._fn(v):
            raise ValueFilteredKeyError(k)
        self._u[k] = v



import typing as ta

from ... import cached
from ... import collections as col
from ... import dataclasses as dc
from ... import lang
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..naming import Naming


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldOptions:
    omit_if: ta.Callable[[ta.Any], bool] | None = None

    default: lang.Maybe[ta.Any] = dc.xfield(default=lang.empty(), check_type=lang.Maybe)

    embed: bool = False

    generic_replace: bool = False

    no_marshal: bool = False
    no_unmarshal: bool = False


DEFAULT_FIELD_OPTIONS = FieldOptions()
FIELD_OPTIONS_KWARGS: frozenset[str] = frozenset(dc.fields_dict(FieldOptions).keys())


@dc.dataclass(frozen=True, kw_only=True)
class FieldMetadata:
    name: str | None = None
    alts: ta.Iterable[str] | None = None

    options: FieldOptions = DEFAULT_FIELD_OPTIONS

    marshaler: Marshaler | None = dc.xfield(None, check_type=(Marshaler, None))
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = dc.xfield(None, check_type=(Unmarshaler, None))
    unmarshaler_factory: UnmarshalerFactory | None = None

    def update(self, **kwargs: ta.Any) -> 'FieldMetadata':
        okw = {k: v for k, v in kwargs.items() if k in FIELD_OPTIONS_KWARGS}
        mkw = {k: v for k, v in kwargs.items() if k not in FIELD_OPTIONS_KWARGS}
        return dc.replace(
            self,
            **(dict(options=dc.replace(self.options, **okw)) if okw else {}),
            **mkw,
        )


@dc.dataclass(frozen=True, kw_only=True)
class ObjectMetadata:
    field_naming: Naming | None = None

    unknown_field: str | None = None
    source_field: str | None = None

    @cached.property
    def specials(self) -> 'ObjectSpecials':
        return ObjectSpecials(
            unknown=self.unknown_field,
            source=self.source_field,
        )

    field_defaults: FieldMetadata = FieldMetadata()

    ignore_unknown: bool = False


@dc.dataclass(frozen=True, kw_only=True)
class ObjectSpecials:
    unknown: str | None = None
    source: str | None = None

    @cached.property
    def set(self) -> frozenset[str]:
        return frozenset(v for v in dc.asdict(self).values() if v is not None)


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldInfo:
    name: str
    type: ta.Any

    marshal_name: str | None
    unmarshal_names: ta.Sequence[str]

    metadata: FieldMetadata = FieldMetadata()

    options: FieldOptions = FieldOptions()


@dc.dataclass(frozen=True)
class FieldInfos:
    lst: ta.Sequence[FieldInfo]

    def __iter__(self) -> ta.Iterator[FieldInfo]:
        return iter(self.lst)

    def __len__(self) -> int:
        return len(self.lst)

    @cached.property
    @dc.init
    def by_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.name, fi) for fi in self), strict=True)

    @cached.property
    @dc.init
    def by_marshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.marshal_name, fi) for fi in self if fi.marshal_name is not None), strict=True)

    @cached.property
    @dc.init
    def by_unmarshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((n, fi) for fi in self for n in fi.unmarshal_names), strict=True)



import typing as ta

from .. import lang
from .codecs import make_bytes_object_codec
from .codecs import make_object_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import cloudpickle
else:
    cloudpickle = lang.proxy_import('cloudpickle')


##


def dump(obj: ta.Any) -> bytes:
    return cloudpickle.dumps(obj)


def load(s: bytes) -> ta.Any:
    return cloudpickle.loads(s)


##


CLOUDPICKLE_CODEC = make_bytes_object_codec('cloudpickle', dump, load)

# @omlish-manifest
_CLOUDPICKLE_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'CLOUDPICKLE_CODEC', CLOUDPICKLE_CODEC)



# type: ignore
# ruff: noqa
# flake8: noqa
# Generated from Protobuf3.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .Protobuf3Parser import Protobuf3Parser
else:
    from Protobuf3Parser import Protobuf3Parser

# This class defines a complete listener for a parse tree produced by Protobuf3Parser.
class Protobuf3Listener(ParseTreeListener):

    # Enter a parse tree produced by Protobuf3Parser#proto.
    def enterProto(self, ctx:Protobuf3Parser.ProtoContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#proto.
    def exitProto(self, ctx:Protobuf3Parser.ProtoContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#syntax.
    def enterSyntax(self, ctx:Protobuf3Parser.SyntaxContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#syntax.
    def exitSyntax(self, ctx:Protobuf3Parser.SyntaxContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#syntaxExtra.
    def enterSyntaxExtra(self, ctx:Protobuf3Parser.SyntaxExtraContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#syntaxExtra.
    def exitSyntaxExtra(self, ctx:Protobuf3Parser.SyntaxExtraContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#importStmt.
    def enterImportStmt(self, ctx:Protobuf3Parser.ImportStmtContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#importStmt.
    def exitImportStmt(self, ctx:Protobuf3Parser.ImportStmtContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#packageStmt.
    def enterPackageStmt(self, ctx:Protobuf3Parser.PackageStmtContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#packageStmt.
    def exitPackageStmt(self, ctx:Protobuf3Parser.PackageStmtContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#option.
    def enterOption(self, ctx:Protobuf3Parser.OptionContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#option.
    def exitOption(self, ctx:Protobuf3Parser.OptionContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#optionName.
    def enterOptionName(self, ctx:Protobuf3Parser.OptionNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#optionName.
    def exitOptionName(self, ctx:Protobuf3Parser.OptionNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#optionBody.
    def enterOptionBody(self, ctx:Protobuf3Parser.OptionBodyContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#optionBody.
    def exitOptionBody(self, ctx:Protobuf3Parser.OptionBodyContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#optionBodyVariable.
    def enterOptionBodyVariable(self, ctx:Protobuf3Parser.OptionBodyVariableContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#optionBodyVariable.
    def exitOptionBodyVariable(self, ctx:Protobuf3Parser.OptionBodyVariableContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#topLevelDef.
    def enterTopLevelDef(self, ctx:Protobuf3Parser.TopLevelDefContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#topLevelDef.
    def exitTopLevelDef(self, ctx:Protobuf3Parser.TopLevelDefContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#message.
    def enterMessage(self, ctx:Protobuf3Parser.MessageContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#message.
    def exitMessage(self, ctx:Protobuf3Parser.MessageContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageBody.
    def enterMessageBody(self, ctx:Protobuf3Parser.MessageBodyContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageBody.
    def exitMessageBody(self, ctx:Protobuf3Parser.MessageBodyContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageBodyContent.
    def enterMessageBodyContent(self, ctx:Protobuf3Parser.MessageBodyContentContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageBodyContent.
    def exitMessageBodyContent(self, ctx:Protobuf3Parser.MessageBodyContentContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#enumDef.
    def enterEnumDef(self, ctx:Protobuf3Parser.EnumDefContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#enumDef.
    def exitEnumDef(self, ctx:Protobuf3Parser.EnumDefContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#enumBody.
    def enterEnumBody(self, ctx:Protobuf3Parser.EnumBodyContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#enumBody.
    def exitEnumBody(self, ctx:Protobuf3Parser.EnumBodyContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#enumField.
    def enterEnumField(self, ctx:Protobuf3Parser.EnumFieldContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#enumField.
    def exitEnumField(self, ctx:Protobuf3Parser.EnumFieldContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#enumValueOption.
    def enterEnumValueOption(self, ctx:Protobuf3Parser.EnumValueOptionContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#enumValueOption.
    def exitEnumValueOption(self, ctx:Protobuf3Parser.EnumValueOptionContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#extend.
    def enterExtend(self, ctx:Protobuf3Parser.ExtendContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#extend.
    def exitExtend(self, ctx:Protobuf3Parser.ExtendContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#service.
    def enterService(self, ctx:Protobuf3Parser.ServiceContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#service.
    def exitService(self, ctx:Protobuf3Parser.ServiceContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#rpc.
    def enterRpc(self, ctx:Protobuf3Parser.RpcContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#rpc.
    def exitRpc(self, ctx:Protobuf3Parser.RpcContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#reserved.
    def enterReserved(self, ctx:Protobuf3Parser.ReservedContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#reserved.
    def exitReserved(self, ctx:Protobuf3Parser.ReservedContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#ranges.
    def enterRanges(self, ctx:Protobuf3Parser.RangesContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#ranges.
    def exitRanges(self, ctx:Protobuf3Parser.RangesContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#rangeRule.
    def enterRangeRule(self, ctx:Protobuf3Parser.RangeRuleContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#rangeRule.
    def exitRangeRule(self, ctx:Protobuf3Parser.RangeRuleContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fieldNames.
    def enterFieldNames(self, ctx:Protobuf3Parser.FieldNamesContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fieldNames.
    def exitFieldNames(self, ctx:Protobuf3Parser.FieldNamesContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#typeRule.
    def enterTypeRule(self, ctx:Protobuf3Parser.TypeRuleContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#typeRule.
    def exitTypeRule(self, ctx:Protobuf3Parser.TypeRuleContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#simpleType.
    def enterSimpleType(self, ctx:Protobuf3Parser.SimpleTypeContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#simpleType.
    def exitSimpleType(self, ctx:Protobuf3Parser.SimpleTypeContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fieldNumber.
    def enterFieldNumber(self, ctx:Protobuf3Parser.FieldNumberContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fieldNumber.
    def exitFieldNumber(self, ctx:Protobuf3Parser.FieldNumberContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#field.
    def enterField(self, ctx:Protobuf3Parser.FieldContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#field.
    def exitField(self, ctx:Protobuf3Parser.FieldContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fieldOptions.
    def enterFieldOptions(self, ctx:Protobuf3Parser.FieldOptionsContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fieldOptions.
    def exitFieldOptions(self, ctx:Protobuf3Parser.FieldOptionsContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fieldOption.
    def enterFieldOption(self, ctx:Protobuf3Parser.FieldOptionContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fieldOption.
    def exitFieldOption(self, ctx:Protobuf3Parser.FieldOptionContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#oneof.
    def enterOneof(self, ctx:Protobuf3Parser.OneofContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#oneof.
    def exitOneof(self, ctx:Protobuf3Parser.OneofContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#oneofField.
    def enterOneofField(self, ctx:Protobuf3Parser.OneofFieldContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#oneofField.
    def exitOneofField(self, ctx:Protobuf3Parser.OneofFieldContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#mapField.
    def enterMapField(self, ctx:Protobuf3Parser.MapFieldContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#mapField.
    def exitMapField(self, ctx:Protobuf3Parser.MapFieldContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#keyType.
    def enterKeyType(self, ctx:Protobuf3Parser.KeyTypeContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#keyType.
    def exitKeyType(self, ctx:Protobuf3Parser.KeyTypeContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#reservedWord.
    def enterReservedWord(self, ctx:Protobuf3Parser.ReservedWordContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#reservedWord.
    def exitReservedWord(self, ctx:Protobuf3Parser.ReservedWordContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fullIdent.
    def enterFullIdent(self, ctx:Protobuf3Parser.FullIdentContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fullIdent.
    def exitFullIdent(self, ctx:Protobuf3Parser.FullIdentContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageName.
    def enterMessageName(self, ctx:Protobuf3Parser.MessageNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageName.
    def exitMessageName(self, ctx:Protobuf3Parser.MessageNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#enumName.
    def enterEnumName(self, ctx:Protobuf3Parser.EnumNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#enumName.
    def exitEnumName(self, ctx:Protobuf3Parser.EnumNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageOrEnumName.
    def enterMessageOrEnumName(self, ctx:Protobuf3Parser.MessageOrEnumNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageOrEnumName.
    def exitMessageOrEnumName(self, ctx:Protobuf3Parser.MessageOrEnumNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#fieldName.
    def enterFieldName(self, ctx:Protobuf3Parser.FieldNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#fieldName.
    def exitFieldName(self, ctx:Protobuf3Parser.FieldNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#oneofName.
    def enterOneofName(self, ctx:Protobuf3Parser.OneofNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#oneofName.
    def exitOneofName(self, ctx:Protobuf3Parser.OneofNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#mapName.
    def enterMapName(self, ctx:Protobuf3Parser.MapNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#mapName.
    def exitMapName(self, ctx:Protobuf3Parser.MapNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#serviceName.
    def enterServiceName(self, ctx:Protobuf3Parser.ServiceNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#serviceName.
    def exitServiceName(self, ctx:Protobuf3Parser.ServiceNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#rpcName.
    def enterRpcName(self, ctx:Protobuf3Parser.RpcNameContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#rpcName.
    def exitRpcName(self, ctx:Protobuf3Parser.RpcNameContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageType.
    def enterMessageType(self, ctx:Protobuf3Parser.MessageTypeContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageType.
    def exitMessageType(self, ctx:Protobuf3Parser.MessageTypeContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#messageOrEnumType.
    def enterMessageOrEnumType(self, ctx:Protobuf3Parser.MessageOrEnumTypeContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#messageOrEnumType.
    def exitMessageOrEnumType(self, ctx:Protobuf3Parser.MessageOrEnumTypeContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#emptyStmt.
    def enterEmptyStmt(self, ctx:Protobuf3Parser.EmptyStmtContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#emptyStmt.
    def exitEmptyStmt(self, ctx:Protobuf3Parser.EmptyStmtContext):
        pass


    # Enter a parse tree produced by Protobuf3Parser#constant.
    def enterConstant(self, ctx:Protobuf3Parser.ConstantContext):
        pass

    # Exit a parse tree produced by Protobuf3Parser#constant.
    def exitConstant(self, ctx:Protobuf3Parser.ConstantContext):
        pass



del Protobuf3Parser


"""
TODO:
 - redacted
 - strongly typed MarshalerFactory base class?
 - strongly typed Composite/Cached Marshaler/Unmarshaler factories - footgun
 - streaming? Start/EndObject, etc..
 - lang.Marker - class name, handle type[Foo]
  - can't disambiguate from str - can't coexist in bare union
 - factories being free MatchFns does more harm than good - in practice these are such big guns you want to write a
   class body if only ceremonially

See:
 - https://github.com/python-attrs/cattrs
 - https://github.com/jcrist/msgspec
 - https://github.com/Fatal1ty/mashumaro
 - https://github.com/Kotlin/kotlinx.serialization/blob/master/docs/serializers.md#custom-serializers

cattrs:
 *

Jackson:
 - USE_ANNOTATIONS
 - AUTO_DETECT_CREATORS
 - AUTO_DETECT_FIELDS
 - AUTO_DETECT_GETTERS
 - AUTO_DETECT_IS_GETTERS
 - AUTO_DETECT_SETTERS
 - REQUIRE_SETTERS_FOR_GETTERS
 - USE_GETTERS_AS_SETTERS
 - INFER_CREATOR_FROM_CONSTRUCTOR_PROPERTIES
 - INFER_PROPERTY_MUTATORS
 - ALLOW_FINAL_FIELDS_AS_MUTATORS
 - ALLOW_VOID_VALUED_PROPERTIES
 - CAN_OVERRIDE_ACCESS_MODIFIERS
 - OVERRIDE_PUBLIC_ACCESS_MODIFIERS
 - SORT_PROPERTIES_ALPHABETICALLY
 - USE_WRAPPER_NAME_AS_PROPERTY_NAME
 - ACCEPT_CASE_INSENSITIVE_ENUMS
 - ACCEPT_CASE_INSENSITIVE_PROPERTIES
 - ACCEPT_CASE_INSENSITIVE_VALUES
 - ALLOW_EXPLICIT_PROPERTY_RENAMING
 - USE_STD_BEAN_NAMING
 - ALLOW_COERCION_OF_SCALARS
 - DEFAULT_VIEW_INCLUSION
 - IGNORE_DUPLICATE_MODULE_REGISTRATIONS
 - IGNORE_MERGE_FOR_UNMERGEABLE
 - USE_BASE_TYPE_AS_DEFAULT_IMPL
 - USE_STATIC_TYPING
 - BLOCK_UNSAFE_POLYMORPHIC_BASE_TYPES

https://github.com/yukinarit/pyserde
 - datatypes
  - typing.Union
  - typing.NewType for primitive types
  - typing.Any
  - typing.Literal
  - typing.Generic
  - typing.ClassVar
 - dataclasses.InitVar
 - Enum and IntEnum
 - pathlib.Path
 - decimal.Decimal
 - uuid.UUID
 - datetime.date, datetime.time, datetime.datetime
 - ipaddress
 - numpy types
 - Class Attributes
 - Field Attributes
 - Decorators
 - Type Check
 - Union Representation
 - Forward reference
 - PEP563 Postponed Evaluation of Annotations
 - PEP585 Type Hinting Generics In Standard Collections
 - PEP604 Allow writing union types as X | Y
 - PEP681 Data Class Transform
 - Case Conversion
 - Rename
 - Alias
 - Skip (de)serialization (skip, skip_if, skip_if_false, skip_if_default)
 - Custom field (de)serializer
 - Custom class (de)serializer
 - Custom global (de)serializer
 - Flatten
"""
import abc
import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from ..funcs import match as mfs
from .errors import UnhandledTypeError
from .factories import RecursiveTypeFactory
from .factories import TypeCacheFactory
from .factories import TypeMapFactory
from .proxy import _Proxy
from .registries import Registry
from .registries import RegistryItem
from .values import Value


##


class Marshaler(lang.Abstract):
    @abc.abstractmethod
    def marshal(self, ctx: 'MarshalContext', o: ta.Any) -> Value:
        raise NotImplementedError


class Unmarshaler(lang.Abstract):
    @abc.abstractmethod
    def unmarshal(self, ctx: 'UnmarshalContext', v: Value) -> ta.Any:
        raise NotImplementedError


##


MarshalerMaker: ta.TypeAlias = mfs.MatchFn[['MarshalContext', rfl.Type], Marshaler]
UnmarshalerMaker: ta.TypeAlias = mfs.MatchFn[['UnmarshalContext', rfl.Type], Unmarshaler]


class MarshalerFactory(lang.Abstract):
    @property
    @abc.abstractmethod
    def make_marshaler(self) -> MarshalerMaker:
        raise NotImplementedError


class UnmarshalerFactory(lang.Abstract):
    @property
    @abc.abstractmethod
    def make_unmarshaler(self) -> UnmarshalerMaker:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class MarshalerFactory_(MarshalerFactory):  # noqa
    fn: MarshalerMaker

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self.fn


@dc.dataclass(frozen=True)
class UnmarshalerFactory_(UnmarshalerFactory):  # noqa
    fn: UnmarshalerMaker

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self.fn


##


class SimpleMarshalerFactory(
    MarshalerFactory,
    MarshalerMaker,
    lang.Abstract,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self


class SimpleUnmarshalerFactory(
    UnmarshalerFactory,
    UnmarshalerMaker,
    lang.Abstract,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self


##


class MarshalerFactoryMatchClass(
    MarshalerFactory,
    mfs.MatchFnClass[['MarshalContext', rfl.Type], Marshaler],
    lang.Abstract,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self


class UnmarshalerFactoryMatchClass(
    UnmarshalerFactory,
    mfs.MatchFnClass[['UnmarshalContext', rfl.Type], Unmarshaler],
    lang.Abstract,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self


##


class MultiMarshalerFactory(MarshalerFactory):
    def __init__(
            self,
            fs: ta.Iterable[MarshalerFactory],
            *,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._fs = list(fs)
        self._mmf: mfs.MultiMatchFn[[MarshalContext, rfl.Type], Marshaler] = mfs.MultiMatchFn(
            [f.make_marshaler for f in self._fs],
            strict=strict,
        )

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._mmf


class MultiUnmarshalerFactory(UnmarshalerFactory):
    def __init__(
            self,
            fs: ta.Iterable[UnmarshalerFactory],
            *,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._fs = list(fs)
        self._mmf: mfs.MultiMatchFn[[UnmarshalContext, rfl.Type], Unmarshaler] = mfs.MultiMatchFn(
            [f.make_unmarshaler for f in self._fs],
            strict=strict,
        )

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._mmf


##


@dc.dataclass(frozen=True)
class TypeMapMarshalerFactory(
    TypeMapFactory['MarshalContext', Marshaler],
    MarshalerFactory,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self


@dc.dataclass(frozen=True)
class TypeMapUnmarshalerFactory(
    TypeMapFactory['UnmarshalContext', Unmarshaler],
    UnmarshalerFactory,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self


# class TypeMapMarshalerFactory(MarshalerFactory):
#     def __init__(self, m: ta.Mapping[rfl.Type, MarshalerFactory]) -> None:
#         super().__init__()
#
#         self._m = m
#         self._tmf: TypeMapFactory['MarshalContext', Marshaler] = TypeMapFactory({
#             t: f.make_marshaler
#             for t, f in m.items()
#         })
#
#     @property
#     def make_marshaler(self) -> MarshalerMaker:
#         return self._tmf


# class TypeMapUnmarshalerFactory(UnmarshalerFactory):
#     def __init__(self, m: ta.Mapping[rfl.Type, UnmarshalerFactory]) -> None:
#         super().__init__()
#
#         self._m = m
#         self._tmf: TypeMapFactory[UnmarshalContext, Unmarshaler] = TypeMapFactory({
#             t: f.make_unmarshaler
#             for t, f in m.items()
#         })
#
#     @property
#     def make_unmarshaler(self) -> UnmarshalerMaker:
#         return self._tmf


##


class TypeCacheMarshalerFactory(MarshalerFactory):
    def __init__(self, f: MarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._tcf: TypeCacheFactory[MarshalContext, Marshaler] = TypeCacheFactory(f.make_marshaler)

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._tcf


class TypeCacheUnmarshalerFactory(UnmarshalerFactory):
    def __init__(self, f: UnmarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._tcf: TypeCacheFactory[UnmarshalContext, Unmarshaler] = TypeCacheFactory(f.make_unmarshaler)

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._tcf


##


@dc.dataclass(frozen=True)
class FuncMarshaler(Marshaler, lang.Final):
    fn: ta.Callable[['MarshalContext', ta.Any], Value]

    def marshal(self, ctx: 'MarshalContext', o: ta.Any) -> Value:
        return self.fn(ctx, o)


@dc.dataclass(frozen=True)
class FuncUnmarshaler(Unmarshaler, lang.Final):
    fn: ta.Callable[['UnmarshalContext', Value], ta.Any]

    def unmarshal(self, ctx: 'UnmarshalContext', v: Value) -> ta.Any:
        return self.fn(ctx, v)


##


class Option:
    pass


##


@dc.dataclass(frozen=True)
class BaseContext(lang.Abstract):
    registry: Registry
    options: col.TypeMap[Option] = col.TypeMap()

    def _reflect(self, o: ta.Any) -> rfl.Type:
        def override(o):
            if (ovr := self.registry.get_of(o, ReflectOverride)):
                return ovr[-1].rty
            return None

        return rfl.Reflector(override=override).type(o)


@dc.dataclass(frozen=True)
class MarshalContext(BaseContext, lang.Final):
    factory: MarshalerFactory | None = None

    def make(self, o: ta.Any) -> Marshaler:
        rty = self._reflect(o)
        try:
            return check.not_none(self.factory).make_marshaler(self, rty)
        except mfs.MatchGuardError:
            raise UnhandledTypeError(rty)  # noqa


@dc.dataclass(frozen=True)
class UnmarshalContext(BaseContext, lang.Final):
    factory: UnmarshalerFactory | None = None

    def make(self, o: ta.Any) -> Unmarshaler:
        rty = self._reflect(o)
        try:
            return check.not_none(self.factory).make_unmarshaler(self, rty)
        except mfs.MatchGuardError:
            raise UnhandledTypeError(rty)  # noqa


##


class _ProxyMarshaler(_Proxy[Marshaler], Marshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self._obj.marshal(ctx, o)


class RecursiveMarshalerFactory(MarshalerFactory, lang.Final):
    def __init__(self, f: MarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._rtf: RecursiveTypeFactory[MarshalContext, Marshaler] = RecursiveTypeFactory(
            self._f.make_marshaler,  # noqa
            _ProxyMarshaler._new,  # noqa
        )

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._rtf


class _ProxyUnmarshaler(_Proxy[Unmarshaler], Unmarshaler):
    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self._obj.unmarshal(ctx, v)


class RecursiveUnmarshalerFactory(UnmarshalerFactory, lang.Final):
    def __init__(self, f: UnmarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._rtf: RecursiveTypeFactory[UnmarshalContext, Unmarshaler] = RecursiveTypeFactory(
            self._f.make_unmarshaler,  # noqa
            _ProxyUnmarshaler._new,  # noqa
        )

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._rtf


##


@dc.dataclass(frozen=True, kw_only=True)
class Override(RegistryItem, lang.Final):
    marshaler: Marshaler | None = dc.xfield(None, check_type=(Marshaler, None))
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = dc.xfield(None, check_type=(Unmarshaler, None))
    unmarshaler_factory: UnmarshalerFactory | None = None


@dc.dataclass(frozen=True)
class ReflectOverride(RegistryItem, lang.Final):
    rty: rfl.Type



# ruff: noqa: UP006 UP007
# @omlish-lite
import collections
import heapq
import typing as ta


T = ta.TypeVar('T')


def distribute_evenly(
        items: ta.Iterable[ta.Tuple[T, float]],
        n_bins: int,
) -> ta.List[ta.List[ta.Tuple[T, float]]]:
    """
    Distribute items into n bins as evenly as possible in terms of total size.
     - Sorting ensures larger items are placed first, preventing large leftover gaps in bins.
     - A min-heap efficiently finds the least loaded bin in O(log n), keeping the distribution balanced.
     - Each item is placed in the lightest bin, preventing a few bins from getting overloaded early.

    :param items: List of tuples (item, size).
    :param n_bins: Number of bins.
    :return: List of n_bins lists, each containing items assigned to that bin.
    """

    # Sort items by size in descending order
    items_sorted = sorted(items, key=lambda x: x[1], reverse=True)

    # Min-heap to track bin loads (size, index)
    bins = [(0, i) for i in range(n_bins)]  # (current size, bin index)
    heapq.heapify(bins)

    # Allocate items to bins
    bin_contents = collections.defaultdict(list)

    for item, size in items_sorted:
        # Get the least loaded bin
        bin_size, bin_index = heapq.heappop(bins)

        # Assign item to this bin
        bin_contents[bin_index].append((item, size))

        # Update bin load and push back to heap
        heapq.heappush(bins, (bin_size + size, bin_index))  # type: ignore

    return [bin_contents[i] for i in range(n_bins)]



import inspect
import sys

from .cached import cached_nullary


@cached_nullary
def is_debugger_attached() -> bool:
    return any(frame[1].endswith('pydevd.py') for frame in inspect.stack())


LITE_REQUIRED_PYTHON_VERSION = (3, 8)


def check_lite_runtime_version() -> None:
    if sys.version_info < LITE_REQUIRED_PYTHON_VERSION:
        raise OSError(f'Requires python {LITE_REQUIRED_PYTHON_VERSION}, got {sys.version_info} from {sys.executable}')  # noqa



import typing as ta

from .. import dataclasses as dc
from .. import lang
from .bindings import Binding
from .elements import Element
from .injector import Injector
from .keys import Key


##


ProvisionListener: ta.TypeAlias = ta.Callable[[
    Injector,
    Key,
    Binding,
    ta.Callable[[], ta.Any],
], ta.Callable[[], ta.Any]]


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class ProvisionListenerBinding(Element, lang.Final):
    listener: ProvisionListener


def bind_provision_listener(l: ProvisionListener) -> Element:
    return ProvisionListenerBinding(l)



import abc
import typing as ta


T = ta.TypeVar('T')


##


_DISABLE_CHECKS = False

_ABSTRACT_METHODS_ATTR = '__abstractmethods__'
_IS_ABSTRACT_METHOD_ATTR = '__isabstractmethod__'
_FORCE_ABSTRACT_ATTR = '__forceabstract__'

_INTERNAL_ABSTRACT_ATTRS = frozenset([_FORCE_ABSTRACT_ATTR])


def make_abstract(obj: T) -> T:
    if callable(obj):
        return ta.cast(T, abc.abstractmethod(obj))
    elif isinstance(obj, property):
        return ta.cast(T, property(
            abc.abstractmethod(obj.fget) if obj.fget is not None else None,
            abc.abstractmethod(obj.fset) if obj.fset is not None else None,
            abc.abstractmethod(obj.fdel) if obj.fdel is not None else None,
        ))
    elif isinstance(obj, (classmethod, staticmethod)):
        return ta.cast(T, type(obj)(abc.abstractmethod(obj.__func__)))
    else:
        return obj


class AbstractTypeError(TypeError):
    pass


class Abstract(abc.ABC):  # noqa
    __slots__ = ()

    def __forceabstract__(self):
        raise TypeError

    setattr(__forceabstract__, _IS_ABSTRACT_METHOD_ATTR, True)

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        if Abstract in cls.__bases__:
            setattr(cls, _FORCE_ABSTRACT_ATTR, getattr(Abstract, _FORCE_ABSTRACT_ATTR))
        else:
            setattr(cls, _FORCE_ABSTRACT_ATTR, False)

        super().__init_subclass__(**kwargs)

        if (
                not _DISABLE_CHECKS and
                Abstract not in cls.__bases__ and
                abc.ABC not in cls.__bases__
        ):
            ams = {a for a, o in cls.__dict__.items() if is_abstract_method(o)}
            seen = set(cls.__dict__)
            for b in cls.__bases__:
                ams.update(set(getattr(b, _ABSTRACT_METHODS_ATTR, [])) - seen)
                seen.update(dir(b))
            if ams:
                raise AbstractTypeError(
                    f'Cannot subclass abstract class {cls.__name__} with abstract methods: '
                    f'{", ".join(map(str, sorted(ams)))}',
                )


def is_abstract_method(obj: ta.Any) -> bool:
    return bool(getattr(obj, _IS_ABSTRACT_METHOD_ATTR, False))


def is_abstract_class(obj: ta.Any) -> bool:
    if bool(getattr(obj, _ABSTRACT_METHODS_ATTR, [])):
        return True
    if isinstance(obj, type):
        if Abstract in obj.__bases__:
            return True
        if (
                Abstract in obj.__mro__
                and getattr(obj.__dict__.get(_FORCE_ABSTRACT_ATTR, None), _IS_ABSTRACT_METHOD_ATTR, False)
        ):
            return True
    return False


def is_abstract(obj: ta.Any) -> bool:
    return is_abstract_method(obj) or is_abstract_class(obj)


def get_abstract_methods(cls: type, *, include_internal: bool = False) -> frozenset[str]:
    ms = frozenset(getattr(cls, _ABSTRACT_METHODS_ATTR))
    if not include_internal:
        ms -= _INTERNAL_ABSTRACT_ATTRS
    return ms


def unabstract_class(
        members: ta.Iterable[str | tuple[str, ta.Any]],
):  # -> ta.Callable[[type[T]], type[T]]:
    def inner(cls):
        if isinstance(members, str):
            raise TypeError(members)

        ams = getattr(cls, _ABSTRACT_METHODS_ATTR)

        names: set[str] = set()
        for m in members:
            if isinstance(m, str):
                if m not in ams:
                    raise NameError(m)
                getattr(cls, m)
                names.add(m)

            elif isinstance(m, tuple):
                name, impl = m
                if name not in ams:
                    raise NameError(name)
                if isinstance(impl, str):
                    impl = getattr(cls, impl)
                setattr(cls, name, impl)
                names.add(name)

        setattr(cls, _ABSTRACT_METHODS_ATTR, ams - names)
        return cls

    return inner



"""
https://swagger.io/specification/
"""
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import marshal as msh
from ...formats import json


##


# https://swagger.io/specification/#security-requirement-object
SecurityRequirement: ta.TypeAlias = ta.Mapping[str, ta.Sequence[str]]


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class OauthFlow:
    """https://swagger.io/specification/#oauth-flow-object"""

    authorization_url: str
    token_url: str
    scopes: ta.Mapping[str, str]
    refresh_ur: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class OauthFlows:
    """https://swagger.io/specification/#oauth-flows-object"""

    implicit: OauthFlow | None = None
    password: OauthFlow | None = None
    client_credentials: OauthFlow | None = None
    authorization_code: OauthFlow | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
@msh.update_fields_metadata(['in_'], name='in')
class SecurityScheme:
    """https://swagger.io/specification/#security-scheme-object"""

    type: str
    scheme: str
    name: str | None = None
    in_: str | None = None
    description: str | None = None
    bearer_format: str | None = None
    flows: OauthFlows | None = None
    open_id_connect_url: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Xml:
    """https://swagger.io/specification/#xml-object"""

    name: str | None = None
    namespace: str | None = None
    prefix: str | None = None
    attribute: bool | None = None
    wrapped: bool | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Discriminator:
    """https://swagger.io/specification/#discriminator-object"""

    property_name: str
    mapping: ta.Mapping[str, str] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class Schema:
    """https://swagger.io/specification/#schema-object"""

    discriminator: Discriminator | None = None
    xml: Xml | None = None
    external_docs: ta.Optional['ExternalDocumentation'] = None
    example: ta.Any | None = None

    # FIXME: HACK: this is a jsonschema lol - these are just hacked on

    type: str | None = None
    format: str | None = None
    one_of: ta.Any = None
    all_of: ta.Any = None
    default: ta.Any = None
    enum: ta.Any = None
    items: ta.Any = None
    required: ta.Any = None
    properties: ta.Any = None
    description: str | None = None
    title: str | None = None
    deprecated: bool | None = None
    nullable: bool | None = None
    additional_properties: ta.Any = None

    #

    x: ta.Mapping[str, ta.Any] | None = None

    @dc.init
    def _check_x(self) -> None:
        for k in self.x or {}:
            check.arg(k.startswith('x-'))


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
@msh.update_fields_metadata(['ref'], name='$ref')
class Reference:
    """https://swagger.io/specification/#reference-object"""

    ref: str
    summary: str | None = None
    description: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Tag:
    """https://swagger.io/specification/#tag-object"""

    name: str
    description: str | None = None
    external_docs: ta.Optional['ExternalDocumentation'] = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
@msh.update_fields_metadata(['common'], embed=True, name='')
class Header:
    """https://swagger.io/specification/#header-object"""

    # TODO: marshal embedding, shared with Parameter
    common: 'ParameterCommon'


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Link:
    """https://swagger.io/specification/#link-object"""

    operation_ref: str | None = None
    operation_id: str | None = None
    parameters: ta.Mapping[str, ta.Any] | None = None
    request_body: ta.Any = None
    description: str | None = None
    server: ta.Optional['Server'] = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Example:
    """https://swagger.io/specification/#example-object"""

    summary: str | None = None
    description: str | None = None
    value: ta.Any | None = None
    external_value: str | None = None


# https://swagger.io/specification/#callback-object
Callback: ta.TypeAlias = ta.Mapping[str, ta.Union['PathItem', Reference]]


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Response:
    """https://swagger.io/specification/#response-object"""

    description: str
    headers: ta.Mapping[str, Header | Reference] | None = None
    content: ta.Mapping[str, 'MediaType'] | None = None
    links: ta.Mapping[str, Link | Reference] | None = None


# https://swagger.io/specification/#responses-object
Responses: ta.TypeAlias = ta.Mapping[str, Response | Reference]


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Encoding:
    """https://swagger.io/specification/#encoding-object"""

    content_type: str | None = None
    headers: ta.Mapping[str, Header | Reference] | None = None
    style: str | None = None
    explode: bool | None = None
    allow_reserved: bool | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class MediaType:
    """https://swagger.io/specification/#media-type-object"""

    schema: Schema | Reference | None = None
    example: ta.Any = None
    examples: ta.Mapping[str, Example | Reference] | None = None
    encoding: ta.Mapping[str, Encoding] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class RequestBody:
    """https://swagger.io/specification/#request-body-object"""

    content: ta.Mapping[str, MediaType]
    description: str | None = None
    required: bool | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class ParameterCommon:
    description: str | None = None
    required: bool | None = None
    deprecated: bool | None = None
    allow_empty_value: bool | None = None

    style: str | None = None
    explode: bool | None = None
    allow_reserved: bool | None = None
    schema: Schema | None = None
    example: ta.Any = None
    examples: ta.Mapping[str, Example | Reference] | None = None

    content: ta.Mapping[str, MediaType] | None = None

    # style: ta.Any = None
    matrix: ta.Any = None
    label: ta.Any = None
    form: ta.Any = None
    simple: ta.Any = None
    space_delimited: ta.Any = None
    pipe_delimited: ta.Any = None
    deep_object: ta.Any = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
@msh.update_fields_metadata(['in_'], name='in')
@msh.update_fields_metadata(['common'], embed=True, name='')
class Parameter:
    """https://swagger.io/specification/#parameter-object"""

    name: str
    in_: str

    common: ParameterCommon


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class ExternalDocumentation:
    """https://swagger.io/specification/#external-documentation-object"""

    url: str
    description: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class Operation:
    """https://swagger.io/specification/#operation-object"""

    tags: ta.Sequence[str] | None = None
    summary: str | None = None
    description: str | None = None
    external_docs: ExternalDocumentation | None = None
    operation_id: str | None = None
    parameters: ta.Sequence[Parameter | Reference] | None = None
    request_body: RequestBody | Reference | None = None
    responses: Responses | None = None
    callbacks: ta.Mapping[str, Callback | Reference] | None = None
    deprecated: bool | None = None
    security: ta.Sequence[SecurityRequirement] | None = None
    servers: ta.Sequence['Server'] | None = None

    #

    x: ta.Mapping[str, ta.Any] | None = None

    @dc.init
    def _check_x(self) -> None:
        for k in self.x or {}:
            check.arg(k.startswith('x-'))


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
@msh.update_fields_metadata(['ref'], name='$ref')
class PathItem:
    """https://swagger.io/specification/#path-item-object"""

    ref: str | None = None
    summary: str | None = None
    description: str | None = None
    get: Operation | None = None
    put: Operation | None = None
    post: Operation | None = None
    delete: Operation | None = None
    options: Operation | None = None
    head: Operation | None = None
    patch: Operation | None = None
    trace: Operation | None = None
    servers: ta.Sequence['Server'] | None = None
    parameters: ta.Sequence[Parameter | Reference] | None = None


# https://swagger.io/specification/#paths-object
Paths: ta.TypeAlias = ta.Mapping[str, PathItem]


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Components:
    """https://swagger.io/specification/#components-object"""

    schemas: ta.Mapping[str, Schema] | None = None
    responses: ta.Mapping[str, Response | Reference] | None = None
    parameters: ta.Mapping[str, Parameter | Reference] | None = None
    examples: ta.Mapping[str, Example | Reference] | None = None
    requestbodies: ta.Mapping[str, RequestBody | Reference] | None = None
    headers: ta.Mapping[str, Header | Reference] | None = None
    security_schemes: ta.Mapping[str, SecurityScheme | Reference] | None = None
    links: ta.Mapping[str, Link | Reference] | None = None
    callbacks: ta.Mapping[str, Callback | Reference] | None = None
    path_items: ta.Mapping[str, PathItem | Reference] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class ServerVariable:
    """https://swagger.io/specification/#server-variable-object"""

    default: str
    enum: ta.Sequence[str] | None = None
    description: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Server:
    """https://swagger.io/specification/#server-object"""

    url: str
    description: str | None = None
    variables: ta.Mapping[str, ServerVariable] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class License:
    """https://swagger.io/specification/#license-object"""

    name: str
    identifier: str | None = None
    url: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Contact:
    """https://swagger.io/specification/#contact-object"""

    name: str | None = None
    url: str | None = None
    email: str | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL)
class Info:
    """https://swagger.io/specification/#info-object"""

    title: str
    version: str
    summary: str | None = None
    description: str | None = None
    terms_of_service: str | None = None
    contact: Contact | None = None
    license: License | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class Openapi:
    """https://swagger.io/specification/#openapi-object"""

    openapi: str
    info: Info
    json_schema_dialect: str | None = None
    servers: ta.Sequence[Server] | None = None
    paths: Paths | None = None
    webhooks: ta.Mapping[str, PathItem | Reference] | None = None
    components: Components | None = None
    security: ta.Sequence[SecurityRequirement] | None = None
    tags: ta.Sequence[Tag] | None = None
    external_docs: ExternalDocumentation | None = None

    #

    x: ta.Mapping[str, ta.Any] | None = None

    @dc.init
    def _check_x(self) -> None:
        for k in self.x or {}:
            check.arg(k.startswith('x-'))


##


def _main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('schema_path')

    args = parser.parse_args()

    import yaml

    with open(args.schema_path) as f:
        doc = yaml.safe_load(f)

    api = msh.unmarshal(doc, Openapi)

    print(json.dumps_pretty(msh.marshal(api)))


if __name__ == '__main__':
    _main()
